import gc
import importlib
import sys
import uuid

from unittest.mock import patch, call, Mock
import pytest
from objgraph import count as count_refs_by_type

from kazoo.testing import KazooTestHarness
from kazoo.exceptions import KazooException
from kazoo.recipe.cache import TreeCache, TreeNode, TreeEvent


class KazooAdaptiveHandlerTestCase(KazooTestHarness):
    HANDLERS = (
        ("kazoo.handlers.gevent", "SequentialGeventHandler"),
        ("kazoo.handlers.eventlet", "SequentialEventletHandler"),
        ("kazoo.handlers.threading", "SequentialThreadingHandler"),
    )

    def setUp(self):
        self.handler = self.choose_an_installed_handler()
        self.setup_zookeeper(handler=self.handler)

    def tearDown(self):
        self.handler = None
        self.teardown_zookeeper()

    def choose_an_installed_handler(self):
        for handler_module, handler_class in self.HANDLERS:
            if (
                handler_module == "kazoo.handlers.gevent"
                and sys.platform == "win32"
            ):
                continue
            try:
                mod = importlib.import_module(handler_module)
                cls = getattr(mod, handler_class)
            except ImportError:
                continue
            else:
                return cls()
        raise ImportError("No available handler")


class KazooTreeCacheTests(KazooAdaptiveHandlerTestCase):
    def setUp(self):
        super(KazooTreeCacheTests, self).setUp()
        self._event_queue = self.client.handler.queue_impl()
        self._error_queue = self.client.handler.queue_impl()
        self.path = None
        self.cache = None

    def tearDown(self):
        if not self._error_queue.empty():
            try:
                raise self._error_queue.get()
            except FakeException:
                pass
        if self.cache is not None:
            self.cache.close()
            self.cache = None
        super(KazooTreeCacheTests, self).tearDown()

    def make_cache(self):
        if self.cache is None:
            self.path = "/" + uuid.uuid4().hex
            self.cache = TreeCache(self.client, self.path)
            self.cache.listen(lambda event: self._event_queue.put(event))
            self.cache.listen_fault(lambda error: self._error_queue.put(error))
            self.cache.start()
        return self.cache

    def wait_cache(self, expect=None, since=None, timeout=10):
        started = since is None
        while True:
            event = self._event_queue.get(timeout=timeout)
            if started:
                if expect is not None:
                    assert event.event_type == expect
                return event
            if event.event_type == since:
                started = True
                if expect is None:
                    return

    def spy_client(self, method_name):
        method = getattr(self.client, method_name)
        return patch.object(self.client, method_name, wraps=method)

    def _wait_gc(self):
        # trigger switching on some coroutine handlers
        self.client.handler.sleep_func(0.1)

        completion_queue = getattr(self.handler, "completion_queue", None)
        if completion_queue is not None:
            while not self.client.handler.completion_queue.empty():
                self.client.handler.sleep_func(0.1)

        for gen in range(3):
            gc.collect(gen)

    def count_tree_node(self):
        # inspect GC and count tree nodes for checking memory leak
        for retry in range(10):
            result = set()
            for _ in range(5):
                self._wait_gc()
                result.add(count_refs_by_type("TreeNode"))
            if len(result) == 1:
                return list(result)[0]
        raise RuntimeError("could not count refs exactly")

    def test_start(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        stat = self.client.exists(self.path)
        assert stat.version == 0

        assert self.cache._state == TreeCache.STATE_STARTED
        assert self.cache._root._state == TreeNode.STATE_LIVE

    def test_start_started(self):
        self.make_cache()
        with pytest.raises(KazooException):
            self.cache.start()

    def test_start_closed(self):
        self.make_cache()
        self.cache.close()
        with pytest.raises(KazooException):
            self.cache.start()

    def test_close(self):
        assert self.count_tree_node() == 0

        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)
        self.client.create(self.path + "/foo/bar/baz", makepath=True)
        for _ in range(3):
            self.wait_cache(TreeEvent.NODE_ADDED)

        # setup stub watchers which are outside of tree cache
        stub_data_watcher = Mock(spec=lambda event: None)
        stub_child_watcher = Mock(spec=lambda event: None)
        self.client.get(self.path + "/foo", stub_data_watcher)
        self.client.get_children(self.path + "/foo", stub_child_watcher)

        # watchers inside tree cache should be here
        root_path = self.client.chroot + self.path
        assert len(self.client._data_watchers[root_path + "/foo"]) == 2
        assert len(self.client._data_watchers[root_path + "/foo/bar"]) == 1
        assert len(self.client._data_watchers[root_path + "/foo/bar/baz"]) == 1
        assert len(self.client._child_watchers[root_path + "/foo"]) == 2
        assert len(self.client._child_watchers[root_path + "/foo/bar"]) == 1
        assert (
            len(self.client._child_watchers[root_path + "/foo/bar/baz"]) == 1
        )

        self.cache.close()

        # nothing should be published since tree closed
        assert self._event_queue.empty()

        # tree should be empty
        assert self.cache._root._children == {}
        assert self.cache._root._data is None
        assert self.cache._state == TreeCache.STATE_CLOSED

        # node state should not be changed
        assert self.cache._root._state != TreeNode.STATE_DEAD

        # watchers should be reset
        assert len(self.client._data_watchers[root_path + "/foo"]) == 1
        assert len(self.client._data_watchers[root_path + "/foo/bar"]) == 0
        assert len(self.client._data_watchers[root_path + "/foo/bar/baz"]) == 0
        assert len(self.client._child_watchers[root_path + "/foo"]) == 1
        assert len(self.client._child_watchers[root_path + "/foo/bar"]) == 0
        assert (
            len(self.client._child_watchers[root_path + "/foo/bar/baz"]) == 0
        )

        # outside watchers should not be deleted
        assert (
            list(self.client._data_watchers[root_path + "/foo"])[0]
            == stub_data_watcher
        )
        assert (
            list(self.client._child_watchers[root_path + "/foo"])[0]
            == stub_child_watcher
        )

        # should not be any leaked memory (tree node) here
        self.cache = None
        assert self.count_tree_node() == 0

    def test_delete_operation(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        assert self.count_tree_node() == 1

        self.client.create(self.path + "/foo/bar/baz", makepath=True)
        for _ in range(3):
            self.wait_cache(TreeEvent.NODE_ADDED)

        self.client.delete(self.path + "/foo", recursive=True)
        for _ in range(3):
            self.wait_cache(TreeEvent.NODE_REMOVED)

        # tree should be empty
        assert self.cache._root._children == {}

        # watchers should be reset
        root_path = self.client.chroot + self.path
        assert self.client._data_watchers[root_path + "/foo"] == set()
        assert self.client._data_watchers[root_path + "/foo/bar"] == set()
        assert self.client._data_watchers[root_path + "/foo/bar/baz"] == set()
        assert self.client._child_watchers[root_path + "/foo"] == set()
        assert self.client._child_watchers[root_path + "/foo/bar"] == set()
        assert self.client._child_watchers[root_path + "/foo/bar/baz"] == set()

        # should not be any leaked memory (tree node) here
        assert self.count_tree_node() == 1

    def test_children_operation(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        self.client.create(self.path + "/test_children", b"test_children_1")
        event = self.wait_cache(TreeEvent.NODE_ADDED)
        assert event.event_type == TreeEvent.NODE_ADDED
        assert event.event_data.path == self.path + "/test_children"
        assert event.event_data.data == b"test_children_1"
        assert event.event_data.stat.version == 0

        self.client.set(self.path + "/test_children", b"test_children_2")
        event = self.wait_cache(TreeEvent.NODE_UPDATED)
        assert event.event_type == TreeEvent.NODE_UPDATED
        assert event.event_data.path == self.path + "/test_children"
        assert event.event_data.data == b"test_children_2"
        assert event.event_data.stat.version == 1

        self.client.delete(self.path + "/test_children")
        event = self.wait_cache(TreeEvent.NODE_REMOVED)
        assert event.event_type == TreeEvent.NODE_REMOVED
        assert event.event_data.path == self.path + "/test_children"
        assert event.event_data.data == b"test_children_2"
        assert event.event_data.stat.version == 1

    def test_subtree_operation(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        self.client.create(self.path + "/foo/bar/baz", makepath=True)
        for relative_path in ("/foo", "/foo/bar", "/foo/bar/baz"):
            event = self.wait_cache(TreeEvent.NODE_ADDED)
            assert event.event_type == TreeEvent.NODE_ADDED
            assert event.event_data.path == self.path + relative_path
            assert event.event_data.data == b""
            assert event.event_data.stat.version == 0

        self.client.delete(self.path + "/foo", recursive=True)
        for relative_path in ("/foo/bar/baz", "/foo/bar", "/foo"):
            event = self.wait_cache(TreeEvent.NODE_REMOVED)
            assert event.event_type == TreeEvent.NODE_REMOVED
            assert event.event_data.path == self.path + relative_path

    def test_get_data(self):
        cache = self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)
        self.client.create(self.path + "/foo/bar/baz", b"@", makepath=True)
        self.wait_cache(TreeEvent.NODE_ADDED)
        self.wait_cache(TreeEvent.NODE_ADDED)
        self.wait_cache(TreeEvent.NODE_ADDED)

        with patch.object(cache, "_client"):  # disable any remote operation
            assert cache.get_data(self.path).data == b""
            assert cache.get_data(self.path).stat.version == 0

            assert cache.get_data(self.path + "/foo").data == b""
            assert cache.get_data(self.path + "/foo").stat.version == 0

            assert cache.get_data(self.path + "/foo/bar").data == b""
            assert cache.get_data(self.path + "/foo/bar").stat.version == 0

            assert cache.get_data(self.path + "/foo/bar/baz").data == b"@"
            assert cache.get_data(self.path + "/foo/bar/baz").stat.version == 0

    def test_get_children(self):
        cache = self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)
        self.client.create(self.path + "/foo/bar/baz", b"@", makepath=True)
        self.wait_cache(TreeEvent.NODE_ADDED)
        self.wait_cache(TreeEvent.NODE_ADDED)
        self.wait_cache(TreeEvent.NODE_ADDED)

        with patch.object(cache, "_client"):  # disable any remote operation
            assert (
                cache.get_children(self.path + "/foo/bar/baz") == frozenset()
            )
            assert cache.get_children(self.path + "/foo/bar") == frozenset(
                ["baz"]
            )
            assert cache.get_children(self.path + "/foo") == frozenset(["bar"])
            assert cache.get_children(self.path) == frozenset(["foo"])

    def test_get_data_out_of_tree(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)
        with pytest.raises(ValueError):
            self.cache.get_data("/out_of_tree")

    def test_get_children_out_of_tree(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)
        with pytest.raises(ValueError):
            self.cache.get_children("/out_of_tree")

    def test_get_data_no_node(self):
        cache = self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        with patch.object(cache, "_client"):  # disable any remote operation
            assert cache.get_data(self.path + "/non_exists") is None

    def test_get_children_no_node(self):
        cache = self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        with patch.object(cache, "_client"):  # disable any remote operation
            assert cache.get_children(self.path + "/non_exists") is None

    def test_session_reconnected(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        self.client.create(self.path + "/foo")
        event = self.wait_cache(TreeEvent.NODE_ADDED)
        assert event.event_data.path == self.path + "/foo"

        with self.spy_client("get_async") as get_data:
            with self.spy_client("get_children_async") as get_children:
                # session suspended
                self.lose_connection(self.client.handler.event_object)
                self.wait_cache(TreeEvent.CONNECTION_SUSPENDED)

                # There are a serial refreshing operation here. But NODE_ADDED
                # events will not be raised because the zxid of nodes are the
                # same during reconnecting.

                # connection restore
                self.wait_cache(TreeEvent.CONNECTION_RECONNECTED)

                # wait for outstanding operations
                while self.cache._outstanding_ops > 0:
                    self.client.handler.sleep_func(0.1)

                # inspect in-memory nodes
                _node_root = self.cache._root
                _node_foo = self.cache._root._children["foo"]

                # make sure that all nodes are refreshed
                get_data.assert_has_calls(
                    [
                        call(self.path, watch=_node_root._process_watch),
                        call(
                            self.path + "/foo", watch=_node_foo._process_watch
                        ),
                    ],
                    any_order=True,
                )
                get_children.assert_has_calls(
                    [
                        call(self.path, watch=_node_root._process_watch),
                        call(
                            self.path + "/foo", watch=_node_foo._process_watch
                        ),
                    ],
                    any_order=True,
                )

    def test_root_recreated(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        # remove root node
        self.client.delete(self.path)
        event = self.wait_cache(TreeEvent.NODE_REMOVED)
        assert event.event_type == TreeEvent.NODE_REMOVED
        assert event.event_data.data == b""
        assert event.event_data.path == self.path
        assert event.event_data.stat.version == 0

        # re-create root node
        self.client.ensure_path(self.path)
        event = self.wait_cache(TreeEvent.NODE_ADDED)
        assert event.event_type == TreeEvent.NODE_ADDED
        assert event.event_data.data == b""
        assert event.event_data.path == self.path
        assert event.event_data.stat.version == 0

        assert self.cache._outstanding_ops >= 0, (
            "unexpected outstanding ops %r" % self.cache._outstanding_ops
        )

    def test_exception_handler(self):
        error_value = FakeException()
        error_handler = Mock()

        with patch.object(TreeNode, "on_deleted") as on_deleted:
            on_deleted.side_effect = [error_value]

            self.make_cache()
            self.cache.listen_fault(error_handler)

            self.cache.close()
            error_handler.assert_called_once_with(error_value)

    def test_exception_suppressed(self):
        self.make_cache()
        self.wait_cache(since=TreeEvent.INITIALIZED)

        # stoke up ConnectionClosedError
        self.client.stop()
        self.client.close()
        self.client.handler.start()  # keep the async completion
        self.wait_cache(since=TreeEvent.CONNECTION_LOST)

        with patch.object(TreeNode, "on_created") as on_created:
            self.cache._root._call_client("exists", "/")
            self.cache._root._call_client("get", "/")
            self.cache._root._call_client("get_children", "/")

            self.wait_cache(since=TreeEvent.INITIALIZED)
            on_created.assert_not_called()
            assert self.cache._outstanding_ops == 0


class FakeException(Exception):
    pass



import sys
from unittest import TestCase

import pytest

from kazoo.protocol import paths


if sys.version_info > (3,):  # pragma: nocover

    def u(s):
        return s

else:  # pragma: nocover

    def u(s):
        return unicode(s, "unicode_escape")  # noqa


class NormPathTestCase(TestCase):
    def test_normpath(self):
        assert paths.normpath("/a/b") == "/a/b"

    def test_normpath_empty(self):
        assert paths.normpath("") == ""

    def test_normpath_unicode(self):
        assert paths.normpath(u("/\xe4/b")) == u("/\xe4/b")

    def test_normpath_dots(self):
        assert paths.normpath("/a./b../c") == "/a./b../c"

    def test_normpath_slash(self):
        assert paths.normpath("/") == "/"

    def test_normpath_multiple_slashes(self):
        assert paths.normpath("//") == "/"
        assert paths.normpath("//a/b") == "/a/b"
        assert paths.normpath("/a//b//") == "/a/b"
        assert paths.normpath("//a////b///c/") == "/a/b/c"

    def test_normpath_relative(self):
        with pytest.raises(ValueError):
            paths.normpath("./a/b")
        with pytest.raises(ValueError):
            paths.normpath("/a/../b")

    def test_normpath_trailing(self):
        assert paths.normpath("/", trailing=True) == "/"


class JoinTestCase(TestCase):
    def test_join(self):
        assert paths.join("/a") == "/a"
        assert paths.join("/a", "b/") == "/a/b/"
        assert paths.join("/a", "b", "c") == "/a/b/c"

    def test_join_empty(self):
        assert paths.join("") == ""
        assert paths.join("", "a", "b") == "a/b"
        assert paths.join("/a", "", "b/", "c") == "/a/b/c"

    def test_join_absolute(self):
        assert paths.join("/a/b", "/c") == "/c"


class IsAbsTestCase(TestCase):
    def test_isabs(self):
        assert paths.isabs("/") is True
        assert paths.isabs("/a") is True
        assert paths.isabs("/a//b/c") is True
        assert paths.isabs("//a/b") is True

    def test_isabs_false(self):
        assert paths.isabs("") is False
        assert paths.isabs("a/") is False
        assert paths.isabs("a/../") is False


class BaseNameTestCase(TestCase):
    def test_basename(self):
        assert paths.basename("") == ""
        assert paths.basename("/") == ""
        assert paths.basename("//a") == "a"
        assert paths.basename("//a/") == ""
        assert paths.basename("/a/b.//c..") == "c.."


class PrefixRootTestCase(TestCase):
    def test_prefix_root(self):
        assert paths._prefix_root("/a/", "b/c") == "/a/b/c"
        assert paths._prefix_root("/a/b", "c/d") == "/a/b/c/d"
        assert paths._prefix_root("/a", "/b/c") == "/a/b/c"
        assert paths._prefix_root("/a", "//b/c.") == "/a/b/c."


class NormRootTestCase(TestCase):
    def test_norm_root(self):
        assert paths._norm_root("") == "/"
        assert paths._norm_root("/") == "/"
        assert paths._norm_root("//a") == "/a"
        assert paths._norm_root("//a./b") == "/a./b"



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging
import os
import time

CI = os.environ.get("CI", False)
CI_ZK_VERSION = CI and os.environ.get("ZOOKEEPER_VERSION", None)
if CI_ZK_VERSION:
    if "-" in CI_ZK_VERSION:
        # Ignore pre-release markers like -alpha
        CI_ZK_VERSION = CI_ZK_VERSION.split("-")[0]
    CI_ZK_VERSION = tuple([int(n) for n in CI_ZK_VERSION.split(".")])


class Handler(logging.Handler):
    def __init__(self, *names, **kw):
        logging.Handler.__init__(self)
        self.names = names
        self.records = []
        self.setLoggerLevel(**kw)

    def setLoggerLevel(self, level=1):
        self.level = level
        self.oldlevels = {}

    def emit(self, record):
        self.records.append(record)

    def clear(self):
        del self.records[:]

    def install(self):
        for name in self.names:
            logger = logging.getLogger(name)
            self.oldlevels[name] = logger.level
            logger.setLevel(self.level)
            logger.addHandler(self)

    def uninstall(self):
        for name in self.names:
            logger = logging.getLogger(name)
            logger.setLevel(self.oldlevels[name])
            logger.removeHandler(self)

    def __str__(self):
        return "\n".join(
            [
                (
                    "%s %s\n  %s"
                    % (
                        record.name,
                        record.levelname,
                        "\n".join(
                            [
                                line
                                for line in record.getMessage().split("\n")
                                if line.strip()
                            ]
                        ),
                    )
                )
                for record in self.records
            ]
        )


class InstalledHandler(Handler):
    def __init__(self, *names, **kw):
        Handler.__init__(self, *names, **kw)
        self.install()


class Wait(object):
    class TimeOutWaitingFor(Exception):
        "A test condition timed out"

    timeout = 9
    wait = 0.01

    def __init__(
        self,
        timeout=None,
        wait=None,
        exception=None,
        getnow=(lambda: time.time),
        getsleep=(lambda: time.sleep),
    ):
        if timeout is not None:
            self.timeout = timeout

        if wait is not None:
            self.wait = wait

        if exception is not None:
            self.TimeOutWaitingFor = exception

        self.getnow = getnow
        self.getsleep = getsleep

    def __call__(self, func=None, timeout=None, wait=None, message=None):
        if func is None:
            return lambda func: self(func, timeout, wait, message)

        if func():
            return

        now = self.getnow()
        sleep = self.getsleep()
        if timeout is None:
            timeout = self.timeout
        if wait is None:
            wait = self.wait
        wait = float(wait)

        deadline = now() + timeout
        while 1:
            sleep(wait)
            if func():
                return
            if now() > deadline:
                raise self.TimeOutWaitingFor(
                    message or func.__doc__ or func.__name__
                )


wait = Wait()



"""Zookeeper Serializers, Deserializers, and NamedTuple objects"""
from collections import namedtuple
import struct

from kazoo.exceptions import EXCEPTIONS
from kazoo.protocol.states import ZnodeStat
from kazoo.security import ACL
from kazoo.security import Id


# Struct objects with formats compiled
bool_struct = struct.Struct("B")
int_struct = struct.Struct("!i")
int_int_struct = struct.Struct("!ii")
int_int_long_struct = struct.Struct("!iiq")

int_long_int_long_struct = struct.Struct("!iqiq")
long_struct = struct.Struct("!q")
multiheader_struct = struct.Struct("!iBi")
reply_header_struct = struct.Struct("!iqi")
stat_struct = struct.Struct("!qqqqiiiqiiq")


def read_string(buffer, offset):
    """Reads an int specified buffer into a string and returns the
    string and the new offset in the buffer"""
    length = int_struct.unpack_from(buffer, offset)[0]
    offset += int_struct.size
    if length < 0:
        return None, offset
    else:
        index = offset
        offset += length
        return buffer[index : index + length].decode("utf-8"), offset


def read_acl(bytes, offset):
    perms = int_struct.unpack_from(bytes, offset)[0]
    offset += int_struct.size
    scheme, offset = read_string(bytes, offset)
    id, offset = read_string(bytes, offset)
    return ACL(perms, Id(scheme, id)), offset


def write_string(bytes):
    if not bytes:
        return int_struct.pack(-1)
    else:
        utf8_str = bytes.encode("utf-8")
        return int_struct.pack(len(utf8_str)) + utf8_str


def write_buffer(bytes):
    if bytes is None:
        return int_struct.pack(-1)
    else:
        return int_struct.pack(len(bytes)) + bytes


def read_buffer(bytes, offset):
    length = int_struct.unpack_from(bytes, offset)[0]
    offset += int_struct.size
    if length < 0:
        return None, offset
    else:
        index = offset
        offset += length
        return bytes[index : index + length], offset


class Close(namedtuple("Close", "")):
    type = -11

    @classmethod
    def serialize(cls):
        return b""


CloseInstance = Close()


class Ping(namedtuple("Ping", "")):
    type = 11

    @classmethod
    def serialize(cls):
        return b""


PingInstance = Ping()


class Connect(
    namedtuple(
        "Connect",
        "protocol_version last_zxid_seen"
        " time_out session_id passwd read_only",
    )
):
    type = None

    def serialize(self):
        b = bytearray()
        b.extend(
            int_long_int_long_struct.pack(
                self.protocol_version,
                self.last_zxid_seen,
                self.time_out,
                self.session_id,
            )
        )
        b.extend(write_buffer(self.passwd))
        b.extend([1 if self.read_only else 0])
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        proto_version, timeout, session_id = int_int_long_struct.unpack_from(
            bytes, offset
        )
        offset += int_int_long_struct.size
        password, offset = read_buffer(bytes, offset)

        try:
            read_only = bool_struct.unpack_from(bytes, offset)[0] == 1
            offset += bool_struct.size
        except struct.error:
            read_only = False
        return (
            cls(proto_version, 0, timeout, session_id, password, read_only),
            offset,
        )


class Create(namedtuple("Create", "path data acl flags")):
    type = 1

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(write_buffer(self.data))
        b.extend(int_struct.pack(len(self.acl)))
        for acl in self.acl:
            b.extend(
                int_struct.pack(acl.perms)
                + write_string(acl.id.scheme)
                + write_string(acl.id.id)
            )
        b.extend(int_struct.pack(self.flags))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        return read_string(bytes, offset)[0]


class Delete(namedtuple("Delete", "path version")):
    type = 2

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(int_struct.pack(self.version))
        return b

    @classmethod
    def deserialize(self, bytes, offset):
        return True


class Exists(namedtuple("Exists", "path watcher")):
    type = 3

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend([1 if self.watcher else 0])
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return stat if stat.czxid != -1 else None


class GetData(namedtuple("GetData", "path watcher")):
    type = 4

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend([1 if self.watcher else 0])
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        data, offset = read_buffer(bytes, offset)
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return data, stat


class SetData(namedtuple("SetData", "path data version")):
    type = 5

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(write_buffer(self.data))
        b.extend(int_struct.pack(self.version))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        return ZnodeStat._make(stat_struct.unpack_from(bytes, offset))


class GetACL(namedtuple("GetACL", "path")):
    type = 6

    def serialize(self):
        return bytearray(write_string(self.path))

    @classmethod
    def deserialize(cls, bytes, offset):
        count = int_struct.unpack_from(bytes, offset)[0]
        offset += int_struct.size
        if count == -1:  # pragma: nocover
            return []

        acls = []
        for c in range(count):
            acl, offset = read_acl(bytes, offset)
            acls.append(acl)
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return acls, stat


class SetACL(namedtuple("SetACL", "path acls version")):
    type = 7

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(int_struct.pack(len(self.acls)))
        for acl in self.acls:
            b.extend(
                int_struct.pack(acl.perms)
                + write_string(acl.id.scheme)
                + write_string(acl.id.id)
            )
        b.extend(int_struct.pack(self.version))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        return ZnodeStat._make(stat_struct.unpack_from(bytes, offset))


class GetChildren(namedtuple("GetChildren", "path watcher")):
    type = 8

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend([1 if self.watcher else 0])
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        count = int_struct.unpack_from(bytes, offset)[0]
        offset += int_struct.size
        if count == -1:  # pragma: nocover
            return []

        children = []
        for c in range(count):
            child, offset = read_string(bytes, offset)
            children.append(child)
        return children


class Sync(namedtuple("Sync", "path")):
    type = 9

    def serialize(self):
        return write_string(self.path)

    @classmethod
    def deserialize(cls, buffer, offset):
        return read_string(buffer, offset)[0]


class GetChildren2(namedtuple("GetChildren2", "path watcher")):
    type = 12

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend([1 if self.watcher else 0])
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        count = int_struct.unpack_from(bytes, offset)[0]
        offset += int_struct.size
        if count == -1:  # pragma: nocover
            return []

        children = []
        for c in range(count):
            child, offset = read_string(bytes, offset)
            children.append(child)
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return children, stat


class CheckVersion(namedtuple("CheckVersion", "path version")):
    type = 13

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(int_struct.pack(self.version))
        return b


class Transaction(namedtuple("Transaction", "operations")):
    type = 14

    def serialize(self):
        b = bytearray()
        for op in self.operations:
            b.extend(
                MultiHeader(op.type, False, -1).serialize() + op.serialize()
            )
        return b + multiheader_struct.pack(-1, True, -1)

    @classmethod
    def deserialize(cls, bytes, offset):
        header = MultiHeader(None, False, None)
        results = []
        response = None
        while not header.done:
            if header.type == Create.type:
                response, offset = read_string(bytes, offset)
            elif header.type == Delete.type:
                response = True
            elif header.type == SetData.type:
                response = ZnodeStat._make(
                    stat_struct.unpack_from(bytes, offset)
                )
                offset += stat_struct.size
            elif header.type == CheckVersion.type:
                response = True
            elif header.type == -1:
                err = int_struct.unpack_from(bytes, offset)[0]
                offset += int_struct.size
                response = EXCEPTIONS[err]()
            if response:
                results.append(response)
            header, offset = MultiHeader.deserialize(bytes, offset)
        return results

    @staticmethod
    def unchroot(client, response):
        resp = []
        for result in response:
            if isinstance(result, str):
                resp.append(client.unchroot(result))
            else:
                resp.append(result)
        return resp


class Create2(namedtuple("Create2", "path data acl flags")):
    type = 15

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.path))
        b.extend(write_buffer(self.data))
        b.extend(int_struct.pack(len(self.acl)))
        for acl in self.acl:
            b.extend(
                int_struct.pack(acl.perms)
                + write_string(acl.id.scheme)
                + write_string(acl.id.id)
            )
        b.extend(int_struct.pack(self.flags))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        path, offset = read_string(bytes, offset)
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return path, stat


class Reconfig(
    namedtuple("Reconfig", "joining leaving new_members config_id")
):
    type = 16

    def serialize(self):
        b = bytearray()
        b.extend(write_string(self.joining))
        b.extend(write_string(self.leaving))
        b.extend(write_string(self.new_members))
        b.extend(long_struct.pack(self.config_id))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        data, offset = read_buffer(bytes, offset)
        stat = ZnodeStat._make(stat_struct.unpack_from(bytes, offset))
        return data, stat


class Auth(namedtuple("Auth", "auth_type scheme auth")):
    type = 100

    def serialize(self):
        return (
            int_struct.pack(self.auth_type)
            + write_string(self.scheme)
            + write_string(self.auth)
        )


class SASL(namedtuple("SASL", "challenge")):
    type = 102

    def serialize(self):
        b = bytearray()
        b.extend(write_buffer(self.challenge))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        challenge, offset = read_buffer(bytes, offset)
        return challenge, offset


class Watch(namedtuple("Watch", "type state path")):
    @classmethod
    def deserialize(cls, bytes, offset):
        """Given bytes and the current bytes offset, return the
        type, state, path, and new offset"""
        type, state = int_int_struct.unpack_from(bytes, offset)
        offset += int_int_struct.size
        path, offset = read_string(bytes, offset)
        return cls(type, state, path), offset


class ReplyHeader(namedtuple("ReplyHeader", "xid, zxid, err")):
    @classmethod
    def deserialize(cls, bytes, offset):
        """Given bytes and the current bytes offset, return a
        :class:`ReplyHeader` instance and the new offset"""
        new_offset = offset + reply_header_struct.size
        return (
            cls._make(reply_header_struct.unpack_from(bytes, offset)),
            new_offset,
        )


class MultiHeader(namedtuple("MultiHeader", "type done err")):
    def serialize(self):
        b = bytearray()
        b.extend(int_struct.pack(self.type))
        b.extend([1 if self.done else 0])
        b.extend(int_struct.pack(self.err))
        return b

    @classmethod
    def deserialize(cls, bytes, offset):
        t, done, err = multiheader_struct.unpack_from(bytes, offset)
        offset += multiheader_struct.size
        return cls(t, done == 1, err), offset



from unittest import TestCase

import pytest


class ExceptionsTestCase(TestCase):
    def _get(self):
        from kazoo import exceptions

        return exceptions

    def test_backwards_alias(self):
        module = self._get()
        assert hasattr(module, "NoNodeException")
        assert module.NoNodeException is module.NoNodeError

    def test_exceptions_code(self):
        module = self._get()
        exc_8 = module.EXCEPTIONS[-8]
        assert isinstance(exc_8(), module.BadArgumentsError)

    def test_invalid_code(self):
        module = self._get()
        with pytest.raises(RuntimeError):
            module.EXCEPTIONS.__getitem__(666)

    def test_exceptions_construction(self):
        module = self._get()
        exc = module.EXCEPTIONS[-101]()
        assert type(exc) is module.NoNodeError
        assert exc.args == ()



import time
import threading
import uuid

import pytest

from kazoo.exceptions import KazooException
from kazoo.protocol.states import EventType
from kazoo.testing import KazooTestCase


class KazooDataWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooDataWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex
        self.client.ensure_path(self.path)

    def test_data_watcher(self):
        update = threading.Event()
        data = [True]

        # Make it a non-existent path
        self.path += "f"

        @self.client.DataWatch(self.path)
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()

        update.wait(10)
        assert data == [None]
        update.clear()

        self.client.create(self.path, b"fred")
        update.wait(10)
        assert data[0] == b"fred"
        update.clear()

    def test_data_watcher_once(self):
        update = threading.Event()
        data = [True]

        # Make it a non-existent path
        self.path += "f"

        dwatcher = self.client.DataWatch(self.path)

        @dwatcher
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()

        update.wait(10)
        assert data == [None]
        update.clear()

        with pytest.raises(KazooException):

            @dwatcher
            def func(d, stat):
                data.pop()

    def test_data_watcher_with_event(self):
        # Test that the data watcher gets passed the event, if it
        # accepts three arguments
        update = threading.Event()
        data = [True]

        # Make it a non-existent path
        self.path += "f"

        @self.client.DataWatch(self.path)
        def changed(d, stat, event):
            data.pop()
            data.append(event)
            update.set()

        update.wait(10)
        assert data == [None]
        update.clear()

        self.client.create(self.path, b"fred")
        update.wait(10)
        assert data[0].type == EventType.CREATED
        update.clear()

    def test_func_style_data_watch(self):
        update = threading.Event()
        data = [True]

        # Make it a non-existent path
        path = self.path + "f"

        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()

        self.client.DataWatch(path, changed)

        update.wait(10)
        assert data == [None]
        update.clear()

        self.client.create(path, b"fred")
        update.wait(10)
        assert data[0] == b"fred"
        update.clear()

    def test_datawatch_across_session_expire(self):
        update = threading.Event()
        data = [True]

        @self.client.DataWatch(self.path)
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()

        update.wait(10)
        assert data == [b""]
        update.clear()

        self.expire_session(threading.Event)
        self.client.retry(self.client.set, self.path, b"fred")
        update.wait(25)
        assert data[0] == b"fred"

    def test_func_stops(self):
        update = threading.Event()
        data = [True]

        self.path += "f"

        fail_through = []

        @self.client.DataWatch(self.path)
        def changed(d, stat):
            data.pop()
            data.append(d)
            update.set()
            if fail_through:
                return False

        update.wait(10)
        assert data == [None]
        update.clear()

        fail_through.append(True)
        self.client.create(self.path, b"fred")
        update.wait(10)
        assert data[0] == b"fred"
        update.clear()

        self.client.set(self.path, b"asdfasdf")
        update.wait(0.2)
        assert data[0] == b"fred"

        d, stat = self.client.get(self.path)
        assert d == b"asdfasdf"

    def test_no_such_node(self):
        args = []

        @self.client.DataWatch("/some/path")
        def changed(d, stat):
            args.extend([d, stat])

        assert args == [None, None]

    def test_no_such_node_for_children_watch(self):
        args = []
        path = self.path + "/test_no_such_node_for_children_watch"
        update = threading.Event()

        def changed(children):
            args.append(children)
            update.set()

        # watch a node which does not exist
        children_watch = self.client.ChildrenWatch(path, changed)
        assert update.is_set() is False
        assert children_watch._stopped is True
        assert args == []

        # watch a node which exists
        self.client.create(path, b"")
        children_watch = self.client.ChildrenWatch(path, changed)
        update.wait(3)
        assert args == [[]]
        update.clear()

        # watch changes
        self.client.create(path + "/fred", b"")
        update.wait(3)
        assert args == [[], ["fred"]]
        update.clear()

        # delete children
        self.client.delete(path + "/fred")
        update.wait(3)
        assert args == [[], ["fred"], []]
        update.clear()

        # delete watching
        self.client.delete(path)

        # a hack for waiting the watcher stop
        for retry in range(5):
            if children_watch._stopped:
                break
            children_watch._run_lock.acquire()
            children_watch._run_lock.release()
            time.sleep(retry / 10.0)

        assert update.is_set() is False
        assert children_watch._stopped is True

    def test_watcher_evaluating_to_false(self):
        class WeirdWatcher(list):
            def __call__(self, *args):
                self.called = True

        watcher = WeirdWatcher()
        self.client.DataWatch(self.path, watcher)
        self.client.set(self.path, b"mwahaha")
        assert watcher.called is True

    def test_watcher_repeat_delete(self):
        a = []
        ev = threading.Event()

        self.client.delete(self.path)

        @self.client.DataWatch(self.path)
        def changed(val, stat):
            a.append(val)
            ev.set()

        assert a == [None]
        ev.wait(10)
        ev.clear()
        self.client.create(self.path, b"blah")
        ev.wait(10)
        assert ev.is_set() is True
        ev.clear()
        assert a == [None, b"blah"]
        self.client.delete(self.path)
        ev.wait(10)
        assert ev.is_set() is True
        ev.clear()
        assert a == [None, b"blah", None]
        self.client.create(self.path, b"blah")
        ev.wait(10)
        assert ev.is_set() is True
        ev.clear()
        assert a == [None, b"blah", None, b"blah"]

    def test_watcher_with_closing(self):
        a = []
        ev = threading.Event()

        self.client.delete(self.path)

        @self.client.DataWatch(self.path)
        def changed(val, stat):
            a.append(val)
            ev.set()

        assert a == [None]

        b = False
        try:
            self.client.stop()
        except:  # noqa
            b = True
        assert b is False


class KazooChildrenWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooChildrenWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex
        self.client.ensure_path(self.path)

    def test_child_watcher(self):
        update = threading.Event()
        all_children = ["fred"]

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait(10)
        assert all_children == []
        update.clear()

        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert all_children == ["smith"]
        update.clear()

        self.client.create(self.path + "/" + "george")
        update.wait(10)
        assert sorted(all_children) == ["george", "smith"]

    def test_child_watcher_once(self):
        update = threading.Event()
        all_children = ["fred"]

        cwatch = self.client.ChildrenWatch(self.path)

        @cwatch
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait(10)
        assert all_children == []
        update.clear()

        with pytest.raises(KazooException):

            @cwatch
            def changed_again(children):
                update.set()

    def test_child_watcher_with_event(self):
        update = threading.Event()
        events = [True]

        @self.client.ChildrenWatch(self.path, send_event=True)
        def changed(children, event):
            events.pop()
            events.append(event)
            update.set()

        update.wait(10)
        assert events == [None]
        update.clear()

        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert events[0].type == EventType.CHILD
        update.clear()

    def test_func_style_child_watcher(self):
        update = threading.Event()
        all_children = ["fred"]

        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        self.client.ChildrenWatch(self.path, changed)

        update.wait(10)
        assert all_children == []
        update.clear()

        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert all_children == ["smith"]
        update.clear()

        self.client.create(self.path + "/" + "george")
        update.wait(10)
        assert sorted(all_children) == ["george", "smith"]

    def test_func_stops(self):
        update = threading.Event()
        all_children = ["fred"]

        fail_through = []

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()
            if fail_through:
                return False

        update.wait(10)
        assert all_children == []
        update.clear()

        fail_through.append(True)
        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert all_children == ["smith"]
        update.clear()

        self.client.create(self.path + "/" + "george")
        update.wait(0.5)
        assert all_children == ["smith"]

    def test_child_watcher_remove_session_watcher(self):
        update = threading.Event()
        all_children = ["fred"]

        fail_through = []

        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()
            if fail_through:
                return False

        children_watch = self.client.ChildrenWatch(self.path, changed)
        session_watcher = children_watch._session_watcher

        update.wait(10)
        assert session_watcher in self.client.state_listeners
        assert all_children == []
        update.clear()

        fail_through.append(True)
        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert session_watcher not in self.client.state_listeners
        assert all_children == ["smith"]
        update.clear()

        self.client.create(self.path + "/" + "george")
        update.wait(10)
        assert session_watcher not in self.client.state_listeners
        assert all_children == ["smith"]

    def test_child_watch_session_loss(self):
        update = threading.Event()
        all_children = ["fred"]

        @self.client.ChildrenWatch(self.path)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait(10)
        assert all_children == []
        update.clear()

        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert all_children == ["smith"]
        update.clear()
        self.expire_session(threading.Event)

        self.client.retry(self.client.create, self.path + "/" + "george")
        update.wait(20)
        assert sorted(all_children) == ["george", "smith"]

    def test_child_stop_on_session_loss(self):
        update = threading.Event()
        all_children = ["fred"]

        @self.client.ChildrenWatch(self.path, allow_session_lost=False)
        def changed(children):
            while all_children:
                all_children.pop()
            all_children.extend(children)
            update.set()

        update.wait(10)
        assert all_children == []
        update.clear()

        self.client.create(self.path + "/" + "smith")
        update.wait(10)
        assert all_children == ["smith"]
        update.clear()
        self.expire_session(threading.Event)

        self.client.retry(self.client.create, self.path + "/" + "george")
        update.wait(4)
        assert update.is_set() is False
        assert all_children == ["smith"]

        children = self.client.get_children(self.path)
        assert sorted(children) == ["george", "smith"]


class KazooPatientChildrenWatcherTests(KazooTestCase):
    def setUp(self):
        super(KazooPatientChildrenWatcherTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def _makeOne(self, *args, **kwargs):
        from kazoo.recipe.watchers import PatientChildrenWatch

        return PatientChildrenWatch(*args, **kwargs)

    def test_watch(self):
        self.client.ensure_path(self.path)
        watcher = self._makeOne(self.client, self.path, 0.1)
        result = watcher.start()
        children, asy = result.get()
        assert len(children) == 0
        assert asy.ready() is False

        self.client.create(self.path + "/" + "fred")
        asy.get(timeout=1)
        assert asy.ready() is True

    def test_exception(self):
        from kazoo.exceptions import NoNodeError

        watcher = self._makeOne(self.client, self.path, 0.1)
        result = watcher.start()

        with pytest.raises(NoNodeError):
            result.get()

    def test_watch_iterations(self):
        self.client.ensure_path(self.path)
        watcher = self._makeOne(self.client, self.path, 0.5)
        result = watcher.start()
        assert result.ready() is False

        time.sleep(0.08)
        self.client.create(self.path + "/" + uuid.uuid4().hex)
        assert result.ready() is False
        time.sleep(0.08)
        assert result.ready() is False
        self.client.create(self.path + "/" + uuid.uuid4().hex)
        time.sleep(0.08)
        assert result.ready() is False

        children, asy = result.get()
        assert len(children) == 2



import uuid

import pytest

from kazoo.testing import KazooTestCase
from kazoo.tests.util import CI_ZK_VERSION


class KazooQueueTests(KazooTestCase):
    def _makeOne(self):
        path = "/" + uuid.uuid4().hex
        return self.client.Queue(path)

    def test_queue_validation(self):
        queue = self._makeOne()
        with pytest.raises(TypeError):
            queue.put({})
        with pytest.raises(TypeError):
            queue.put(b"one", b"100")
        with pytest.raises(TypeError):
            queue.put(b"one", 10.0)
        with pytest.raises(ValueError):
            queue.put(b"one", -100)
        with pytest.raises(ValueError):
            queue.put(b"one", 100000)

    def test_empty_queue(self):
        queue = self._makeOne()
        assert len(queue) == 0
        assert queue.get() is None
        assert len(queue) == 0

    def test_queue(self):
        queue = self._makeOne()
        queue.put(b"one")
        queue.put(b"two")
        queue.put(b"three")
        assert len(queue) == 3

        assert queue.get() == b"one"
        assert queue.get() == b"two"
        assert queue.get() == b"three"
        assert len(queue) == 0

    def test_priority(self):
        queue = self._makeOne()
        queue.put(b"four", priority=101)
        queue.put(b"one", priority=0)
        queue.put(b"two", priority=0)
        queue.put(b"three", priority=10)

        assert queue.get() == b"one"
        assert queue.get() == b"two"
        assert queue.get() == b"three"
        assert queue.get() == b"four"


class KazooLockingQueueTests(KazooTestCase):
    def setUp(self):
        KazooTestCase.setUp(self)
        skip = False
        if CI_ZK_VERSION and CI_ZK_VERSION < (3, 4):
            skip = True
        elif CI_ZK_VERSION and CI_ZK_VERSION >= (3, 4):
            skip = False
        else:
            ver = self.client.server_version()
            if ver[1] < 4:
                skip = True
        if skip:
            pytest.skip("Must use Zookeeper 3.4 or above")

    def _makeOne(self):
        path = "/" + uuid.uuid4().hex
        return self.client.LockingQueue(path)

    def test_queue_validation(self):
        queue = self._makeOne()
        with pytest.raises(TypeError):
            queue.put({})
        with pytest.raises(TypeError):
            queue.put(b"one", b"100")
        with pytest.raises(TypeError):
            queue.put(b"one", 10.0)
        with pytest.raises(ValueError):
            queue.put(b"one", -100)
        with pytest.raises(ValueError):
            queue.put(b"one", 100000)
        with pytest.raises(TypeError):
            queue.put_all({})
        with pytest.raises(TypeError):
            queue.put_all([{}])
        with pytest.raises(TypeError):
            queue.put_all([b"one"], b"100")
        with pytest.raises(TypeError):
            queue.put_all([b"one"], 10.0)
        with pytest.raises(ValueError):
            queue.put_all([b"one"], -100)
        with pytest.raises(ValueError):
            queue.put_all([b"one"], 100000)

    def test_empty_queue(self):
        queue = self._makeOne()
        assert len(queue) == 0
        assert queue.get(0) is None
        assert len(queue) == 0

    def test_queue(self):
        queue = self._makeOne()
        queue.put(b"one")
        queue.put_all([b"two", b"three"])
        assert len(queue) == 3

        assert not queue.consume()
        assert not queue.holds_lock()
        assert queue.get(1) == b"one"
        assert queue.holds_lock()
        # Without consuming, should return the same element
        assert queue.get(1) == b"one"
        assert queue.consume()
        assert not queue.holds_lock()
        assert queue.get(1) == b"two"
        assert queue.holds_lock()
        assert queue.consume()
        assert not queue.holds_lock()
        assert queue.get(1) == b"three"
        assert queue.holds_lock()
        assert queue.consume()
        assert not queue.holds_lock()
        assert not queue.consume()
        assert len(queue) == 0

    def test_consume(self):
        queue = self._makeOne()

        queue.put(b"one")
        assert not queue.consume()
        queue.get(0.1)
        assert queue.consume()
        assert not queue.consume()

    def test_release(self):
        queue = self._makeOne()

        queue.put(b"one")
        assert queue.get(1) == b"one"
        assert queue.holds_lock()
        assert queue.release()
        assert not queue.holds_lock()
        assert queue.get(1) == b"one"
        assert queue.consume()
        assert not queue.release()
        assert len(queue) == 0

    def test_holds_lock(self):
        queue = self._makeOne()

        assert not queue.holds_lock()
        queue.put(b"one")
        queue.get(0.1)
        assert queue.holds_lock()
        queue.consume()
        assert not queue.holds_lock()

    def test_priority(self):
        queue = self._makeOne()
        queue.put(b"four", priority=101)
        queue.put(b"one", priority=0)
        queue.put(b"two", priority=0)
        queue.put(b"three", priority=10)

        assert queue.get(1) == b"one"
        assert queue.consume()
        assert queue.get(1) == b"two"
        assert queue.consume()
        assert queue.get(1) == b"three"
        assert queue.consume()
        assert queue.get(1) == b"four"
        assert queue.consume()

    def test_concurrent_execution(self):
        queue = self._makeOne()
        value1 = []
        value2 = []
        value3 = []
        event1 = self.client.handler.event_object()
        event2 = self.client.handler.event_object()
        event3 = self.client.handler.event_object()

        def get_concurrently(value, event):
            q = self.client.LockingQueue(queue.path)
            value.append(q.get(0.1))
            event.set()

        self.client.handler.spawn(get_concurrently, value1, event1)
        self.client.handler.spawn(get_concurrently, value2, event2)
        self.client.handler.spawn(get_concurrently, value3, event3)
        queue.put(b"one")
        event1.wait(0.2)
        event2.wait(0.2)
        event3.wait(0.2)

        result = value1 + value2 + value3
        assert result.count(b"one") == 1
        assert result.count(None) == 2



import urllib.parse


def collect_hosts(hosts):
    """
    Collect a set of hosts and an optional chroot from
    a string or a list of strings.
    """
    if isinstance(hosts, list):
        if hosts[-1].strip().startswith("/"):
            host_ports, chroot = hosts[:-1], hosts[-1]
        else:
            host_ports, chroot = hosts, None
    else:
        host_ports, chroot = hosts.partition("/")[::2]
        host_ports = host_ports.split(",")
        chroot = "/" + chroot if chroot else None

    result = []
    for host_port in host_ports:
        # put all complexity of dealing with
        # IPv4 & IPv6 address:port on the urlsplit
        res = urllib.parse.urlsplit("xxx://" + host_port)
        host = res.hostname
        if host is None:
            raise ValueError("bad hostname")
        port = int(res.port) if res.port else 2181
        result.append((host.strip(), port))

    return result, chroot



import uuid
import sys
import threading

import pytest

from kazoo.testing import KazooTestCase
from kazoo.tests.util import wait


class UniqueError(Exception):
    """Error raised only by test leader function"""


class KazooElectionTests(KazooTestCase):
    def setUp(self):
        super(KazooElectionTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

        self.condition = threading.Condition()

        # election contenders set these when elected. The exit event is set by
        # the test to make the leader exit.
        self.leader_id = None
        self.exit_event = None

        # tests set this before the event to make the leader raise an error
        self.raise_exception = False

        # set by a worker thread when an unexpected error is hit.
        # better way to do this?
        self.thread_exc_info = None

    def _spawn_contender(self, contender_id, election):
        thread = threading.Thread(
            target=self._election_thread, args=(contender_id, election)
        )
        thread.daemon = True
        thread.start()
        return thread

    def _election_thread(self, contender_id, election):
        try:
            election.run(self._leader_func, contender_id)
        except UniqueError:
            if not self.raise_exception:
                self.thread_exc_info = sys.exc_info()
        except Exception:
            self.thread_exc_info = sys.exc_info()
        else:
            if self.raise_exception:
                e = Exception("expected leader func to raise exception")
                self.thread_exc_info = (Exception, e, None)

    def _leader_func(self, name):
        exit_event = threading.Event()
        with self.condition:
            self.exit_event = exit_event
            self.leader_id = name
            self.condition.notify_all()

        exit_event.wait(45)
        if self.raise_exception:
            raise UniqueError("expected error in the leader function")

    def _check_thread_error(self):
        if self.thread_exc_info:
            t, o, tb = self.thread_exc_info
            raise t(o)

    def test_election(self):
        elections = {}
        threads = {}
        for _ in range(3):
            contender = "c" + uuid.uuid4().hex
            elections[contender] = self.client.Election(self.path, contender)
            threads[contender] = self._spawn_contender(
                contender, elections[contender]
            )

        # wait for a leader to be elected
        times = 0
        with self.condition:
            while not self.leader_id:
                self.condition.wait(5)
                times += 1
                if times > 5:
                    raise Exception(
                        "Still not a leader: lid: %s", self.leader_id
                    )

        election = self.client.Election(self.path)

        # make sure all contenders are in the pool
        wait(lambda: len(election.contenders()) == len(elections))
        contenders = election.contenders()

        assert set(contenders) == set(elections.keys())

        # first one in list should be leader
        first_leader = contenders[0]
        assert first_leader == self.leader_id

        # tell second one to cancel election. should never get elected.
        elections[contenders[1]].cancel()

        # make leader exit. third contender should be elected.
        self.exit_event.set()
        with self.condition:
            while self.leader_id == first_leader:
                self.condition.wait(45)
        assert self.leader_id == contenders[2]
        self._check_thread_error()

        # make first contender re-enter the race
        threads[first_leader].join()
        threads[first_leader] = self._spawn_contender(
            first_leader, elections[first_leader]
        )

        # contender set should now be the current leader plus the first leader
        wait(lambda: len(election.contenders()) == 2)
        contenders = election.contenders()
        assert set(contenders), set([self.leader_id == first_leader])

        # make current leader raise an exception. first should be reelected
        self.raise_exception = True
        self.exit_event.set()
        with self.condition:
            while self.leader_id != first_leader:
                self.condition.wait(45)
        assert self.leader_id == first_leader
        self._check_thread_error()

        self.exit_event.set()
        for thread in threads.values():
            thread.join()
        self._check_thread_error()

    def test_bad_func(self):
        election = self.client.Election(self.path)
        with pytest.raises(ValueError):
            election.run("not a callable")



import collections
import threading
import unittest
from unittest.mock import MagicMock
import uuid

import pytest

from kazoo.exceptions import CancelledError
from kazoo.exceptions import LockTimeout
from kazoo.exceptions import NoNodeError
from kazoo.recipe.lock import Lock
from kazoo.testing import KazooTestCase
from kazoo.tests import util as test_util


class SleepBarrier(object):
    """A crappy spinning barrier."""

    def __init__(self, wait_for, sleep_func):
        self._wait_for = wait_for
        self._arrived = collections.deque()
        self._sleep_func = sleep_func

    def __enter__(self):
        self._arrived.append(threading.current_thread())
        return self

    def __exit__(self, type, value, traceback):
        try:
            self._arrived.remove(threading.current_thread())
        except ValueError:
            pass

    def wait(self):
        while len(self._arrived) < self._wait_for:
            self._sleep_func(0.001)


class KazooLockTests(KazooTestCase):
    thread_count = 20

    def __init__(self, *args, **kw):
        super(KazooLockTests, self).__init__(*args, **kw)
        self.threads_made = []

    def tearDown(self):
        super(KazooLockTests, self).tearDown()
        while self.threads_made:
            t = self.threads_made.pop()
            t.join()

    @staticmethod
    def make_condition():
        return threading.Condition()

    @staticmethod
    def make_event():
        return threading.Event()

    def make_thread(self, *args, **kwargs):
        t = threading.Thread(*args, **kwargs)
        t.daemon = True
        self.threads_made.append(t)
        return t

    @staticmethod
    def make_wait():
        return test_util.Wait()

    def setUp(self):
        super(KazooLockTests, self).setUp()
        self.lockpath = "/" + uuid.uuid4().hex
        self.condition = self.make_condition()
        self.released = self.make_event()
        self.active_thread = None
        self.cancelled_threads = []

    def _thread_lock_acquire_til_event(self, name, lock, event):
        try:
            with lock:
                with self.condition:
                    assert self.active_thread is None
                    self.active_thread = name
                    self.condition.notify_all()

                event.wait()

                with self.condition:
                    assert self.active_thread == name
                    self.active_thread = None
                    self.condition.notify_all()
            self.released.set()
        except CancelledError:
            with self.condition:
                self.cancelled_threads.append(name)
                self.condition.notify_all()

    def test_lock_one(self):
        lock_name = uuid.uuid4().hex
        lock = self.client.Lock(self.lockpath, lock_name)
        event = self.make_event()
        thread = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=(lock_name, lock, event),
        )
        thread.start()

        lock2_name = uuid.uuid4().hex
        anotherlock = self.client.Lock(self.lockpath, lock2_name)

        # wait for any contender to show up on the lock
        wait = self.make_wait()
        wait(anotherlock.contenders)
        assert anotherlock.contenders() == [lock_name]

        with self.condition:
            while self.active_thread != lock_name:
                self.condition.wait()

        # release the lock
        event.set()

        with self.condition:
            while self.active_thread:
                self.condition.wait()
        self.released.wait()
        thread.join()

    def test_lock(self):
        threads = []
        names = ["contender" + str(i) for i in range(5)]

        contender_bits = {}

        for name in names:
            ev = self.make_event()
            lock = self.client.Lock(self.lockpath, name)
            thread = self.make_thread(
                target=self._thread_lock_acquire_til_event,
                args=(name, lock, ev),
            )
            contender_bits[name] = (thread, ev)
            threads.append(thread)

        # acquire the lock ourselves first to make the others line up
        lock = self.client.Lock(self.lockpath, "test")
        lock.acquire()

        for thread in threads:
            thread.start()

        # wait for everyone to line up on the lock
        wait = self.make_wait()
        wait(lambda: len(lock.contenders()) == 6)
        contenders = lock.contenders()

        assert contenders[0] == "test"
        contenders = contenders[1:]
        remaining = list(contenders)

        # release the lock and contenders should claim it in order
        lock.release()

        for contender in contenders:
            thread, event = contender_bits[contender]

            with self.condition:
                while not self.active_thread:
                    self.condition.wait()
                assert self.active_thread == contender

            assert lock.contenders() == remaining
            remaining = remaining[1:]

            event.set()

            with self.condition:
                while self.active_thread:
                    self.condition.wait()
        for thread in threads:
            thread.join()

    def test_lock_reconnect(self):
        event = self.make_event()
        other_lock = self.client.Lock(self.lockpath, "contender")
        thread = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("contender", other_lock, event),
        )

        # acquire the lock ourselves first to make the contender line up
        lock = self.client.Lock(self.lockpath, "test")
        lock.acquire()

        thread.start()
        # wait for the contender to line up on the lock
        wait = self.make_wait()
        wait(lambda: len(lock.contenders()) == 2)
        assert lock.contenders() == ["test", "contender"]

        self.expire_session(self.make_event)

        lock.release()

        with self.condition:
            while not self.active_thread:
                self.condition.wait()
            assert self.active_thread == "contender"

        event.set()
        thread.join()

    def test_lock_non_blocking(self):
        lock_name = uuid.uuid4().hex
        lock = self.client.Lock(self.lockpath, lock_name)
        event = self.make_event()

        thread = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=(lock_name, lock, event),
        )
        thread.start()

        lock1 = self.client.Lock(self.lockpath, lock_name)

        # wait for the thread to acquire the lock
        with self.condition:
            if not self.active_thread:
                self.condition.wait(5)

        assert not lock1.acquire(blocking=False)
        assert lock.contenders() == [lock_name]  # just one - itself

        event.set()
        thread.join()

    def test_lock_fail_first_call(self):
        event1 = self.make_event()
        lock1 = self.client.Lock(self.lockpath, "one")
        thread1 = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("one", lock1, event1),
        )
        thread1.start()

        # wait for this thread to acquire the lock
        with self.condition:
            if not self.active_thread:
                self.condition.wait(5)
                assert self.active_thread == "one"
        assert lock1.contenders() == ["one"]
        event1.set()
        thread1.join()

    def test_lock_cancel(self):
        event1 = self.make_event()
        lock1 = self.client.Lock(self.lockpath, "one")
        thread1 = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("one", lock1, event1),
        )
        thread1.start()

        # wait for this thread to acquire the lock
        with self.condition:
            if not self.active_thread:
                self.condition.wait(5)
                assert self.active_thread == "one"

        client2 = self._get_client()
        client2.start()
        event2 = self.make_event()
        lock2 = client2.Lock(self.lockpath, "two")
        thread2 = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("two", lock2, event2),
        )
        thread2.start()

        # this one should block in acquire. check that it is a contender
        wait = self.make_wait()
        wait(lambda: len(lock2.contenders()) > 1)
        assert lock2.contenders() == ["one", "two"]

        lock2.cancel()
        with self.condition:
            if "two" not in self.cancelled_threads:
                self.condition.wait()
                assert "two" in self.cancelled_threads

        assert lock2.contenders() == ["one"]

        thread2.join()
        event1.set()
        thread1.join()
        client2.stop()

    def test_lock_no_double_calls(self):
        lock1 = self.client.Lock(self.lockpath, "one")
        lock1.acquire()
        assert lock1.is_acquired is True
        assert lock1.acquire(timeout=0.5) is False
        assert lock1.is_acquired is True
        lock1.release()
        assert lock1.is_acquired is False

    def test_lock_same_thread_no_block(self):
        lock = self.client.Lock(self.lockpath, "one")
        gotten = lock.acquire(blocking=False)
        assert gotten is True
        assert lock.is_acquired is True
        gotten = lock.acquire(blocking=False)
        assert gotten is False

    def test_lock_many_threads_no_block(self):
        lock = self.client.Lock(self.lockpath, "one")
        attempts = collections.deque()

        def _acquire():
            attempts.append(int(lock.acquire(blocking=False)))

        threads = []
        for _i in range(0, self.thread_count):
            t = self.make_thread(target=_acquire)
            threads.append(t)
            t.start()

        while threads:
            t = threads.pop()
            t.join()

        assert sum(list(attempts)) == 1

    def test_lock_many_threads(self):
        sleep_func = self.client.handler.sleep_func
        lock = self.client.Lock(self.lockpath, "one")
        acquires = collections.deque()
        differences = collections.deque()
        barrier = SleepBarrier(self.thread_count, sleep_func)

        def _acquire():
            # Wait until all threads are ready to go...
            with barrier as b:
                b.wait()
                with lock:
                    # Ensure that no two threads enter here and cause the
                    # count to differ by more than one, do this by recording
                    # the count that was captured and examining it post run.
                    starting_count = len(acquires)
                    acquires.append(1)
                    sleep_func(0.01)
                    end_count = len(acquires)
                    differences.append(end_count - starting_count)

        threads = []
        for _i in range(0, self.thread_count):
            t = self.make_thread(target=_acquire)
            threads.append(t)
            t.start()

        while threads:
            t = threads.pop()
            t.join()

        assert len(acquires) == self.thread_count
        assert list(differences) == [1] * self.thread_count

    def test_lock_reacquire(self):
        lock = self.client.Lock(self.lockpath, "one")
        lock.acquire()
        lock.release()
        lock.acquire()
        lock.release()

    def test_lock_ephemeral(self):
        client1 = self._get_client()
        client1.start()
        lock = client1.Lock(self.lockpath, "ephemeral")
        lock.acquire(ephemeral=False)
        znode = self.lockpath + "/" + lock.node
        client1.stop()
        try:
            self.client.get(znode)
        except NoNodeError:
            self.fail("NoNodeError raised unexpectedly!")

    def test_lock_timeout(self):
        timeout = 3
        e = self.make_event()
        started = self.make_event()

        # In the background thread, acquire the lock and wait thrice the time
        # that the main thread is going to wait to acquire the lock.
        lock1 = self.client.Lock(self.lockpath, "one")

        def _thread(lock, event, timeout):
            with lock:
                started.set()
                event.wait(timeout)
                if not event.is_set():
                    # Eventually fail to avoid hanging the tests
                    self.fail("lock2 never timed out")

        t = self.make_thread(target=_thread, args=(lock1, e, timeout * 3))
        t.start()

        # Start the main thread's kazoo client and try to acquire the lock
        # but give up after `timeout` seconds
        client2 = self._get_client()
        client2.start()
        started.wait(5)
        assert started.is_set() is True
        lock2 = client2.Lock(self.lockpath, "two")
        try:
            lock2.acquire(timeout=timeout)
        except LockTimeout:
            # A timeout is the behavior we're expecting, since the background
            # thread should still be holding onto the lock
            pass
        else:
            self.fail("Main thread unexpectedly acquired the lock")
        finally:
            # Cleanup
            e.set()
            t.join()
            client2.stop()

    def test_read_lock(self):
        # Test that we can obtain a read lock
        lock = self.client.ReadLock(self.lockpath, "reader one")
        gotten = lock.acquire(blocking=False)
        assert gotten is True
        assert lock.is_acquired is True
        # and that it's still not reentrant.
        gotten = lock.acquire(blocking=False)
        assert gotten is False
        # Test that a second client we can share the same read lock
        client2 = self._get_client()
        client2.start()
        lock2 = client2.ReadLock(self.lockpath, "reader two")
        gotten = lock2.acquire(blocking=False)
        assert gotten is True
        assert lock2.is_acquired is True
        gotten = lock2.acquire(blocking=False)
        assert gotten is False
        # Test that a writer is unable to share it
        client3 = self._get_client()
        client3.start()
        lock3 = client3.WriteLock(self.lockpath, "writer")
        gotten = lock3.acquire(blocking=False)
        assert gotten is False

    def test_write_lock(self):
        # Test that we can obtain a write lock
        lock = self.client.WriteLock(self.lockpath, "writer")
        gotten = lock.acquire(blocking=False)
        assert gotten is True
        assert lock.is_acquired is True
        gotten = lock.acquire(blocking=False)
        assert gotten is False

        # Test that we are unable to obtain a read lock while the
        # write lock is held.
        client2 = self._get_client()
        client2.start()
        lock2 = client2.ReadLock(self.lockpath, "reader")
        gotten = lock2.acquire(blocking=False)
        assert gotten is False

    def test_rw_lock(self):
        reader_event = self.make_event()
        reader_lock = self.client.ReadLock(self.lockpath, "reader")
        reader_thread = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("reader", reader_lock, reader_event),
        )

        writer_event = self.make_event()
        writer_lock = self.client.WriteLock(self.lockpath, "writer")
        writer_thread = self.make_thread(
            target=self._thread_lock_acquire_til_event,
            args=("writer", writer_lock, writer_event),
        )

        # acquire a write lock ourselves first to make the others line up
        lock = self.client.WriteLock(self.lockpath, "test")
        lock.acquire()

        reader_thread.start()
        writer_thread.start()

        # wait for everyone to line up on the lock
        wait = self.make_wait()
        wait(lambda: len(lock.contenders()) == 3)
        contenders = lock.contenders()

        assert contenders[0] == "test"
        remaining = contenders[1:]

        # release the lock and contenders should claim it in order
        lock.release()

        contender_bits = {
            "reader": (reader_thread, reader_event),
            "writer": (writer_thread, writer_event),
        }

        for contender in ("reader", "writer"):
            thread, event = contender_bits[contender]

            with self.condition:
                while not self.active_thread:
                    self.condition.wait()
                assert self.active_thread == contender

            assert lock.contenders() == remaining
            remaining = remaining[1:]

            event.set()

            with self.condition:
                while self.active_thread:
                    self.condition.wait()

        reader_thread.join()
        writer_thread.join()


class TestSemaphore(KazooTestCase):
    def __init__(self, *args, **kw):
        super(TestSemaphore, self).__init__(*args, **kw)
        self.threads_made = []

    def tearDown(self):
        super(TestSemaphore, self).tearDown()
        while self.threads_made:
            t = self.threads_made.pop()
            t.join()

    @staticmethod
    def make_condition():
        return threading.Condition()

    @staticmethod
    def make_event():
        return threading.Event()

    def make_thread(self, *args, **kwargs):
        t = threading.Thread(*args, **kwargs)
        t.daemon = True
        self.threads_made.append(t)
        return t

    def setUp(self):
        super(TestSemaphore, self).setUp()
        self.lockpath = "/" + uuid.uuid4().hex
        self.condition = self.make_condition()
        self.released = self.make_event()
        self.active_thread = None
        self.cancelled_threads = []

    def test_basic(self):
        sem1 = self.client.Semaphore(self.lockpath)
        sem1.acquire()
        sem1.release()

    def test_lock_one(self):
        sem1 = self.client.Semaphore(self.lockpath, max_leases=1)
        sem2 = self.client.Semaphore(self.lockpath, max_leases=1)
        started = self.make_event()
        event = self.make_event()

        sem1.acquire()

        def sema_one():
            started.set()
            with sem2:
                event.set()

        thread = self.make_thread(target=sema_one, args=())
        thread.start()
        started.wait(10)

        assert event.is_set() is False

        sem1.release()
        event.wait(10)
        assert event.is_set() is True
        thread.join()

    def test_non_blocking(self):
        sem1 = self.client.Semaphore(
            self.lockpath, identifier="sem1", max_leases=2
        )
        sem2 = self.client.Semaphore(
            self.lockpath, identifier="sem2", max_leases=2
        )
        sem3 = self.client.Semaphore(
            self.lockpath, identifier="sem3", max_leases=2
        )

        sem1.acquire()
        sem2.acquire()
        assert not sem3.acquire(blocking=False)
        assert set(sem1.lease_holders()) == set(["sem1", "sem2"])
        sem2.release()
        # the next line isn't required, but avoids timing issues in tests
        sem3.acquire()
        assert set(sem1.lease_holders()) == set(["sem1", "sem3"])
        sem1.release()
        sem3.release()

    def test_non_blocking_release(self):
        sem1 = self.client.Semaphore(
            self.lockpath, identifier="sem1", max_leases=1
        )
        sem2 = self.client.Semaphore(
            self.lockpath, identifier="sem2", max_leases=1
        )
        sem1.acquire()
        sem2.acquire(blocking=False)

        # make sure there's no shutdown / cleanup error
        sem1.release()
        sem2.release()

    def test_holders(self):
        started = self.make_event()
        event = self.make_event()

        def sema_one():
            with self.client.Semaphore(self.lockpath, "fred", max_leases=1):
                started.set()
                event.wait()

        thread = self.make_thread(target=sema_one, args=())
        thread.start()
        started.wait()
        sem1 = self.client.Semaphore(self.lockpath)
        holders = sem1.lease_holders()
        assert holders == ["fred"]
        event.set()
        thread.join()

    def test_semaphore_cancel(self):
        sem1 = self.client.Semaphore(self.lockpath, "fred", max_leases=1)
        sem2 = self.client.Semaphore(self.lockpath, "george", max_leases=1)
        sem1.acquire()
        started = self.make_event()
        event = self.make_event()

        def sema_one():
            started.set()
            try:
                with sem2:
                    started.set()
            except CancelledError:
                event.set()

        thread = self.make_thread(target=sema_one, args=())
        thread.start()
        started.wait()
        assert sem1.lease_holders() == ["fred"]
        assert not event.is_set()
        sem2.cancel()
        event.wait()
        assert event.is_set()
        thread.join()

    def test_multiple_acquire_and_release(self):
        sem1 = self.client.Semaphore(self.lockpath, "fred", max_leases=1)
        sem1.acquire()
        sem1.acquire()

        assert sem1.release()
        assert not sem1.release()

    def test_handle_session_loss(self):
        expire_semaphore = self.client.Semaphore(
            self.lockpath, "fred", max_leases=1
        )

        client = self._get_client()
        client.start()
        lh_semaphore = client.Semaphore(self.lockpath, "george", max_leases=1)
        lh_semaphore.acquire()

        started = self.make_event()
        event = self.make_event()
        event2 = self.make_event()

        def sema_one():
            started.set()
            with expire_semaphore:
                event.set()
                event2.wait()

        thread1 = self.make_thread(target=sema_one, args=())
        thread1.start()

        started.wait()
        assert lh_semaphore.lease_holders() == ["george"]

        # Fired in a separate thread to make sure we can see the effect
        expired = self.make_event()

        def expire():
            self.expire_session(self.make_event)
            expired.set()

        thread2 = self.make_thread(target=expire, args=())
        thread2.start()
        expire_semaphore.wake_event.wait()
        expired.wait()

        lh_semaphore.release()
        client.stop()

        event.wait(15)
        assert expire_semaphore.lease_holders() == ["fred"]
        event2.set()

        for t in (thread1, thread2):
            t.join()

    def test_inconsistent_max_leases(self):
        sem1 = self.client.Semaphore(self.lockpath, max_leases=1)
        sem2 = self.client.Semaphore(self.lockpath, max_leases=2)

        sem1.acquire()
        with pytest.raises(ValueError):
            sem2.acquire()

    def test_inconsistent_max_leases_other_data(self):
        sem1 = self.client.Semaphore(self.lockpath, max_leases=1)
        sem2 = self.client.Semaphore(self.lockpath, max_leases=2)

        self.client.ensure_path(self.lockpath)
        self.client.set(self.lockpath, b"a$")

        sem1.acquire()
        # sem2 thinks it's ok to have two lease holders
        assert sem2.acquire(blocking=False)

    def test_reacquire(self):
        lock = self.client.Semaphore(self.lockpath)
        lock.acquire()
        lock.release()
        lock.acquire()
        lock.release()

    def test_acquire_after_cancelled(self):
        lock = self.client.Semaphore(self.lockpath)
        assert lock.acquire() is True
        assert lock.release() is True
        lock.cancel()
        assert lock.cancelled is True
        assert lock.acquire() is True

    def test_timeout(self):
        timeout = 3
        e = self.make_event()
        started = self.make_event()

        # In the background thread, acquire the lock and wait thrice the time
        # that the main thread is going to wait to acquire the lock.
        sem1 = self.client.Semaphore(self.lockpath, "one")

        def _thread(sem, event, timeout):
            with sem:
                started.set()
                event.wait(timeout)
                if not event.is_set():
                    # Eventually fail to avoid hanging the tests
                    self.fail("sem2 never timed out")

        t = self.make_thread(target=_thread, args=(sem1, e, timeout * 3))
        t.start()

        # Start the main thread's kazoo client and try to acquire the lock
        # but give up after `timeout` seconds
        client2 = self._get_client()
        client2.start()
        started.wait(5)
        assert started.is_set() is True
        sem2 = client2.Semaphore(self.lockpath, "two")
        try:
            sem2.acquire(timeout=timeout)
        except LockTimeout:
            # A timeout is the behavior we're expecting, since the background
            # thread will still be holding onto the lock
            e.set()
        finally:
            # Cleanup
            t.join()
            client2.stop()


class TestSequence(unittest.TestCase):
    def test_get_predecessor(self):
        """Validate selection of predecessors."""
        goLock = "_c_8eb60557ba51e0da67eefc47467d3f34-lock-0000000031"
        pyLock = "514e5a831836450cb1a56c741e990fd8__lock__0000000032"
        children = ["hello", goLock, "world", pyLock]
        client = MagicMock()
        client.get_children.return_value = children
        lock = Lock(client, "test")
        assert lock._get_predecessor(pyLock) is None

    def test_get_predecessor_go(self):
        """Test selection of predecessor when instructed to consider go-zk
        locks.
        """
        goLock = "_c_8eb60557ba51e0da67eefc47467d3f34-lock-0000000031"
        pyLock = "514e5a831836450cb1a56c741e990fd8__lock__0000000032"
        children = ["hello", goLock, "world", pyLock]
        client = MagicMock()
        client.get_children.return_value = children
        lock = Lock(client, "test", extra_lock_patterns=["-lock-"])
        assert lock._get_predecessor(pyLock) == goLock



import datetime
import uuid

from kazoo.recipe.lease import NonBlockingLease
from kazoo.recipe.lease import MultiNonBlockingLease

from kazoo.testing import KazooTestCase


class MockClock(object):
    def __init__(self, epoch=0):
        self.epoch = epoch

    def forward(self, seconds):
        self.epoch += seconds

    def __call__(self):
        return datetime.datetime.utcfromtimestamp(self.epoch)


class KazooLeaseTests(KazooTestCase):
    def setUp(self):
        super(KazooLeaseTests, self).setUp()
        self.client2 = self._get_client(timeout=0.8)
        self.client2.start()
        self.client3 = self._get_client(timeout=0.8)
        self.client3.start()
        self.path = "/" + uuid.uuid4().hex
        self.clock = MockClock(10)

    def tearDown(self):
        for cl in [self.client2, self.client3]:
            if cl.connected:
                cl.stop()
                cl.close()
        del self.client2
        del self.client3


class NonBlockingLeaseTests(KazooLeaseTests):
    def test_renew(self):
        # Use client convenience method here to test it at least once.  Use
        # class directly in
        # other tests in order to get better IDE support.
        lease = self.client.NonBlockingLease(
            self.path, datetime.timedelta(seconds=3), utcnow=self.clock
        )
        assert lease
        assert lease.obtained is True

        self.clock.forward(2)
        renewed_lease = self.client.NonBlockingLease(
            self.path, datetime.timedelta(seconds=3), utcnow=self.clock
        )
        assert renewed_lease

    def test_busy(self):
        lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert lease

        self.clock.forward(2)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert not foreigner_lease
        assert foreigner_lease.obtained is False

    def test_overtake(self):
        lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert lease

        self.clock.forward(4)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert foreigner_lease

    def test_renew_no_overtake(self):
        lease = self.client.NonBlockingLease(
            self.path, datetime.timedelta(seconds=3), utcnow=self.clock
        )
        assert lease
        assert lease.obtained is True

        self.clock.forward(2)
        renewed_lease = self.client.NonBlockingLease(
            self.path, datetime.timedelta(seconds=3), utcnow=self.clock
        )
        assert renewed_lease

        self.clock.forward(2)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert not foreigner_lease

    def test_overtaker_renews(self):
        lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert lease

        self.clock.forward(4)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert foreigner_lease

        self.clock.forward(2)
        foreigner_renew = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert foreigner_renew

    def test_overtake_refuse_first(self):
        lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert lease

        self.clock.forward(4)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert foreigner_lease

        self.clock.forward(2)
        first_again_lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert not first_again_lease

    def test_old_version(self):
        # Skip to a future version
        NonBlockingLease._version += 1
        lease = NonBlockingLease(
            self.client,
            self.path,
            datetime.timedelta(seconds=3),
            utcnow=self.clock,
        )
        assert lease

        # Then back to today.
        NonBlockingLease._version -= 1
        self.clock.forward(4)
        foreigner_lease = NonBlockingLease(
            self.client2,
            self.path,
            datetime.timedelta(seconds=3),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        # Since a newer version wrote the lease file, the lease is not taken.
        assert not foreigner_lease


class MultiNonBlockingLeaseTest(KazooLeaseTests):
    def test_1_renew(self):
        ls = self.client.MultiNonBlockingLease(
            1, self.path, datetime.timedelta(seconds=4), utcnow=self.clock
        )
        assert ls
        self.clock.forward(2)
        ls2 = MultiNonBlockingLease(
            self.client,
            1,
            self.path,
            datetime.timedelta(seconds=4),
            utcnow=self.clock,
        )
        assert ls2

    def test_1_reject(self):
        ls = MultiNonBlockingLease(
            self.client,
            1,
            self.path,
            datetime.timedelta(seconds=4),
            utcnow=self.clock,
        )
        assert ls
        self.clock.forward(2)
        ls2 = MultiNonBlockingLease(
            self.client2,
            1,
            self.path,
            datetime.timedelta(seconds=4),
            identifier="some.other.host",
            utcnow=self.clock,
        )
        assert not ls2

    def test_2_renew(self):
        ls = MultiNonBlockingLease(
            self.client,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            utcnow=self.clock,
        )
        assert ls
        self.clock.forward(2)
        ls2 = MultiNonBlockingLease(
            self.client2,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            identifier="host2",
            utcnow=self.clock,
        )
        assert ls2
        self.clock.forward(2)
        ls3 = MultiNonBlockingLease(
            self.client,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            utcnow=self.clock,
        )
        assert ls3
        self.clock.forward(2)
        ls4 = MultiNonBlockingLease(
            self.client2,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            identifier="host2",
            utcnow=self.clock,
        )
        assert ls4

    def test_2_reject(self):
        ls = MultiNonBlockingLease(
            self.client,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            utcnow=self.clock,
        )
        assert ls
        self.clock.forward(2)
        ls2 = MultiNonBlockingLease(
            self.client2,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            identifier="host2",
            utcnow=self.clock,
        )
        assert ls2
        self.clock.forward(2)
        ls3 = MultiNonBlockingLease(
            self.client3,
            2,
            self.path,
            datetime.timedelta(seconds=7),
            identifier="host3",
            utcnow=self.clock,
        )
        assert not ls3

    def test_2_handover(self):
        ls = MultiNonBlockingLease(
            self.client,
            2,
            self.path,
            datetime.timedelta(seconds=4),
            utcnow=self.clock,
        )
        assert ls
        self.clock.forward(2)
        ls2 = MultiNonBlockingLease(
            self.client2,
            2,
            self.path,
            datetime.timedelta(seconds=4),
            identifier="host2",
            utcnow=self.clock,
        )
        assert ls2
        self.clock.forward(3)
        ls3 = MultiNonBlockingLease(
            self.client3,
            2,
            self.path,
            datetime.timedelta(seconds=4),
            identifier="host3",
            utcnow=self.clock,
        )
        assert ls3
        self.clock.forward(2)
        ls4 = MultiNonBlockingLease(
            self.client,
            2,
            self.path,
            datetime.timedelta(seconds=4),
            utcnow=self.clock,
        )
        assert ls4



import unittest

from kazoo.security import Permissions


class TestACL(unittest.TestCase):
    def _makeOne(self, *args, **kwargs):
        from kazoo.security import make_acl

        return make_acl(*args, **kwargs)

    def test_read_acl(self):
        acl = self._makeOne("digest", ":", read=True)
        assert acl.perms & Permissions.READ == Permissions.READ

    def test_all_perms(self):
        acl = self._makeOne(
            "digest",
            ":",
            read=True,
            write=True,
            create=True,
            delete=True,
            admin=True,
        )
        for perm in [
            Permissions.READ,
            Permissions.CREATE,
            Permissions.WRITE,
            Permissions.DELETE,
            Permissions.ADMIN,
        ]:
            assert acl.perms & perm == perm

    def test_perm_listing(self):
        from kazoo.security import ACL

        f = ACL(15, "fred")
        assert "READ" in f.acl_list
        assert "WRITE" in f.acl_list
        assert "CREATE" in f.acl_list
        assert "DELETE" in f.acl_list

        f = ACL(16, "fred")
        assert "ADMIN" in f.acl_list

        f = ACL(31, "george")
        assert "ALL" in f.acl_list

    def test_perm_repr(self):
        from kazoo.security import ACL

        f = ACL(16, "fred")
        assert "ACL(perms=16, acl_list=['ADMIN']" in repr(f)



from unittest import TestCase

from kazoo.hosts import collect_hosts


class HostsTestCase(TestCase):
    def test_ipv4(self):
        hosts, chroot = collect_hosts(
            "127.0.0.1:2181, 192.168.1.2:2181, \
                                       132.254.111.10:2181"
        )
        assert hosts == [
            ("127.0.0.1", 2181),
            ("192.168.1.2", 2181),
            ("132.254.111.10", 2181),
        ]
        assert chroot is None

        hosts, chroot = collect_hosts(
            ["127.0.0.1:2181", "192.168.1.2:2181", "132.254.111.10:2181"]
        )
        assert hosts == [
            ("127.0.0.1", 2181),
            ("192.168.1.2", 2181),
            ("132.254.111.10", 2181),
        ]
        assert chroot is None

    def test_ipv6(self):
        hosts, chroot = collect_hosts("[fe80::200:5aee:feaa:20a2]:2181")
        assert hosts == [("fe80::200:5aee:feaa:20a2", 2181)]
        assert chroot is None

        hosts, chroot = collect_hosts(["[fe80::200:5aee:feaa:20a2]:2181"])
        assert hosts == [("fe80::200:5aee:feaa:20a2", 2181)]
        assert chroot is None

    def test_hosts_list(self):
        hosts, chroot = collect_hosts("zk01:2181, zk02:2181, zk03:2181")
        expected1 = [("zk01", 2181), ("zk02", 2181), ("zk03", 2181)]
        assert hosts == expected1
        assert chroot is None

        hosts, chroot = collect_hosts(["zk01:2181", "zk02:2181", "zk03:2181"])
        assert hosts == expected1
        assert chroot is None

        expected2 = "/test"
        hosts, chroot = collect_hosts("zk01:2181, zk02:2181, zk03:2181/test")
        assert hosts == expected1
        assert chroot == expected2

        hosts, chroot = collect_hosts(
            ["zk01:2181", "zk02:2181", "zk03:2181", "/test"]
        )
        assert hosts == expected1
        assert chroot == expected2



"""Zookeeper lease implementations

:Maintainer: Lars Albertsson <lars.albertsson@gmail.com>
:Maintainer: Jyrki Pulliainen <jyrki@spotify.com>
:Status: Beta

"""
import datetime
import json
import socket

from kazoo.exceptions import CancelledError


class NonBlockingLease(object):
    """Exclusive lease that does not block.

    An exclusive lease ensures that only one client at a time owns the lease.
    The client may renew the lease without losing it by obtaining a new lease
    with the same path and same identity.  The lease object evaluates to True
    if the lease was obtained.

    A common use case is a situation where a task should only run on a single
    host.  In this case, the clients that did not obtain the lease should exit
    without performing the protected task.

    The lease stores time stamps using client clocks, and will therefore only
    work if client clocks are roughly synchronised.  It uses UTC, and works
    across time zones and daylight savings.

    Example usage: with a :class:`~kazoo.client.KazooClient` instance::

        zk = KazooClient()
        zk.start()
        # Hold lease over an hour in order to keep job on same machine,
        # with failover if it dies.
        lease = zk.NonBlockingLease(
            "/db_leases/hourly_cleanup", datetime.timedelta(minutes = 70),
            identifier = "DB hourly cleanup on " + socket.gethostname())
        if lease:
            do_hourly_database_cleanup()
    """

    # Bump when storage format changes
    _version = 1
    _date_format = "%Y-%m-%dT%H:%M:%S"
    _byte_encoding = "utf-8"

    def __init__(
        self,
        client,
        path,
        duration,
        identifier=None,
        utcnow=datetime.datetime.utcnow,
    ):
        """Create a non-blocking lease.

        :param client: A :class:`~kazoo.client.KazooClient` instance.
        :param path: The lease path to use.
        :param duration: Duration during which the lease is reserved.  A
                         :class:`~datetime.timedelta` instance.
        :param identifier: Unique name to use for this lease holder. Reuse in
                           order to renew the lease. Defaults to
                           :meth:`socket.gethostname()`.
        :param utcnow: Clock function, by default returning
                       :meth:`datetime.datetime.utcnow()`. Used for testing.

        """
        ident = identifier or socket.gethostname()
        self.obtained = False
        self._attempt_obtaining(client, path, duration, ident, utcnow)

    def _attempt_obtaining(self, client, path, duration, ident, utcnow):
        client.ensure_path(path)
        holder_path = path + "/lease_holder"
        lock = client.Lock(path, ident)
        try:
            with lock:
                now = utcnow()
                if client.exists(holder_path):
                    raw, _ = client.get(holder_path)
                    data = self._decode(raw)
                    if data["version"] != self._version:
                        # We need an upgrade, let someone else take the lease
                        return
                    current_end = datetime.datetime.strptime(
                        data["end"], self._date_format
                    )
                    if data["holder"] != ident and now < current_end:
                        # Another client is still holding the lease
                        return
                    client.delete(holder_path)
                end_lease = (now + duration).strftime(self._date_format)
                new_data = {
                    "version": self._version,
                    "holder": ident,
                    "end": end_lease,
                }
                client.create(holder_path, self._encode(new_data))
                self.obtained = True

        except CancelledError:
            pass

    def _encode(self, data_dict):
        return json.dumps(data_dict).encode(self._byte_encoding)

    def _decode(self, raw):
        return json.loads(raw.decode(self._byte_encoding))

    # Python 2.x
    def __nonzero__(self):
        return self.obtained

    # Python 3.x
    def __bool__(self):
        return self.obtained


class MultiNonBlockingLease(object):
    """Exclusive lease for multiple clients.

    This type of lease is useful when a limited set of hosts should run a
    particular task. It will attempt to obtain leases trying a sequence of
    ZooKeeper lease paths.

    :param client: A :class:`~kazoo.client.KazooClient` instance.
    :param count: Number of host leases allowed.
    :param path: ZooKeeper path under which lease files are stored.
    :param duration: Duration during which the lease is reserved.  A
                     :class:`~datetime.timedelta` instance.
    :param identifier: Unique name to use for this lease holder. Reuse in order
                       to renew the lease.
           Defaults do :meth:`socket.gethostname()`.
    :param utcnow: Clock function, by default returning
                   :meth:`datetime.datetime.utcnow()`.  Used for testing.

    """

    def __init__(
        self,
        client,
        count,
        path,
        duration,
        identifier=None,
        utcnow=datetime.datetime.utcnow,
    ):
        self.obtained = False
        for num in range(count):
            ls = NonBlockingLease(
                client,
                "%s/%d" % (path, num),
                duration,
                identifier=identifier,
                utcnow=utcnow,
            )
            if ls:
                self.obtained = True
                break

    # Python 2.x
    def __nonzero__(self):
        return self.obtained

    # Python 3.x
    def __bool__(self):
        return self.obtained



"""ZooKeeper Leader Elections

:Maintainer: None
:Status: Unknown

"""
from kazoo.exceptions import CancelledError


class Election(object):
    """Kazoo Basic Leader Election

    Example usage with a :class:`~kazoo.client.KazooClient` instance::

        zk = KazooClient()
        zk.start()
        election = zk.Election("/electionpath", "my-identifier")

        # blocks until the election is won, then calls
        # my_leader_function()
        election.run(my_leader_function)

    """

    def __init__(self, client, path, identifier=None):
        """Create a Kazoo Leader Election

        :param client: A :class:`~kazoo.client.KazooClient` instance.
        :param path: The election path to use.
        :param identifier: Name to use for this lock contender. This
                           can be useful for querying to see who the
                           current lock contenders are.

        """
        self.lock = client.Lock(path, identifier)

    def run(self, func, *args, **kwargs):
        """Contend for the leadership

        This call will block until either this contender is cancelled
        or this contender wins the election and the provided leadership
        function subsequently returns or fails.

        :param func: A function to be called if/when the election is
                     won.
        :param args: Arguments to leadership function.
        :param kwargs: Keyword arguments to leadership function.

        """
        if not callable(func):
            raise ValueError("leader function is not callable")

        try:
            with self.lock:
                func(*args, **kwargs)

        except CancelledError:
            pass

    def cancel(self):
        """Cancel participation in the election

        .. note::

            If this contender has already been elected leader, this
            method will not interrupt the leadership function.

        """
        self.lock.cancel()

    def contenders(self):
        """Return an ordered list of the current contenders in the
        election

        .. note::

            If the contenders did not set an identifier, it will appear
            as a blank string.

        """
        return self.lock.contenders()



import unittest
import sys

import pytest

from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError
from kazoo.protocol.states import Callback
from kazoo.testing import KazooTestCase
from kazoo.tests import test_client


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
class TestGeventHandler(unittest.TestCase):
    def setUp(self):
        try:
            import gevent  # NOQA
        except ImportError:
            pytest.skip("gevent not available.")

    def _makeOne(self, *args):
        from kazoo.handlers.gevent import SequentialGeventHandler

        return SequentialGeventHandler(*args)

    def _getAsync(self, *args):
        from kazoo.handlers.gevent import AsyncResult

        return AsyncResult

    def _getEvent(self):
        from gevent.event import Event

        return Event

    def test_proper_threading(self):
        h = self._makeOne()
        h.start()
        assert isinstance(h.event_object(), self._getEvent())

    def test_matching_async(self):
        h = self._makeOne()
        h.start()
        async_handler = self._getAsync()
        assert isinstance(h.async_result(), async_handler)

    def test_exception_raising(self):
        h = self._makeOne()

        with pytest.raises(h.timeout_exception):
            raise h.timeout_exception("This is a timeout")

    def test_exception_in_queue(self):
        h = self._makeOne()
        h.start()
        ev = self._getEvent()()

        def func():
            ev.set()
            raise ValueError("bang")

        call1 = Callback("completion", func, ())
        h.dispatch_callback(call1)
        ev.wait()

    def test_queue_empty_exception(self):
        from gevent.queue import Empty

        h = self._makeOne()
        h.start()
        ev = self._getEvent()()

        def func():
            ev.set()
            raise Empty()

        call1 = Callback("completion", func, ())
        h.dispatch_callback(call1)
        ev.wait()


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
class TestBasicGeventClient(KazooTestCase):
    def setUp(self):
        try:
            import gevent  # NOQA
        except ImportError:
            pytest.skip("gevent not available.")
        KazooTestCase.setUp(self)

    def _makeOne(self, *args):
        from kazoo.handlers.gevent import SequentialGeventHandler

        return SequentialGeventHandler(*args)

    def _getEvent(self):
        from gevent.event import Event

        return Event

    def test_start(self):
        client = self._get_client(handler=self._makeOne())
        client.start()
        assert client.state == "CONNECTED"
        client.stop()

    def test_start_stop_double(self):
        client = self._get_client(handler=self._makeOne())
        client.start()
        assert client.state == "CONNECTED"
        client.handler.start()
        client.handler.stop()
        client.stop()

    def test_basic_commands(self):
        client = self._get_client(handler=self._makeOne())
        client.start()
        assert client.state == "CONNECTED"
        client.create("/anode", b"fred")
        assert client.get("/anode")[0] == b"fred"
        assert client.delete("/anode")
        assert client.exists("/anode") is None
        client.stop()

    def test_failures(self):
        client = self._get_client(handler=self._makeOne())
        client.start()
        with pytest.raises(NoNodeError):
            client.get("/none")
        client.stop()

    def test_data_watcher(self):
        client = self._get_client(handler=self._makeOne())
        client.start()
        client.ensure_path("/some/node")
        ev = self._getEvent()()

        @client.DataWatch("/some/node")
        def changed(d, stat):
            ev.set()

        ev.wait()
        ev.clear()
        client.set("/some/node", b"newvalue")
        ev.wait()
        client.stop()

    def test_huge_file_descriptor(self):
        import resource
        from gevent import socket
        from kazoo.handlers.utils import create_tcp_socket

        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (4096, 4096))
        except (ValueError, resource.error):
            self.skipTest("couldnt raise fd limit high enough")
        fd = 0
        socks = []
        while fd < 4000:
            sock = create_tcp_socket(socket)
            fd = sock.fileno()
            socks.append(sock)
        h = self._makeOne()
        h.start()
        h.select(socks, [], [], 0)
        h.stop()
        for sock in socks:
            sock.close()


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
class TestGeventClient(test_client.TestClient):
    def setUp(self):
        try:
            import gevent  # NOQA
        except ImportError:
            pytest.skip("gevent not available.")
        KazooTestCase.setUp(self)

    def _makeOne(self, *args):
        from kazoo.handlers.gevent import SequentialGeventHandler

        return SequentialGeventHandler(*args)

    def _get_client(self, **kwargs):
        kwargs["handler"] = self._makeOne()
        return KazooClient(self.hosts, **kwargs)



"""TreeCache

:Maintainer: Jiangge Zhang <tonyseek@gmail.com>
:Maintainer: Haochuan Guo <guohaochuan@gmail.com>
:Maintainer: Tianwen Zhang <mail2tevin@gmail.com>
:Status: Alpha

A port of the Apache Curator's TreeCache recipe. It builds an in-memory cache
of a subtree in ZooKeeper and keeps it up-to-date.

See also: http://curator.apache.org/curator-recipes/tree-cache.html
"""
from __future__ import absolute_import

import contextlib
import functools
import logging
import operator

from kazoo.exceptions import NoNodeError, KazooException
from kazoo.protocol.paths import _prefix_root, join as kazoo_join
from kazoo.protocol.states import KazooState, EventType

logger = logging.getLogger(__name__)


class TreeCache(object):
    """The cache of a ZooKeeper subtree.

    :param client: A :class:`~kazoo.client.KazooClient` instance.
    :param path: The root path of subtree.
    """

    STATE_LATENT = 0
    STATE_STARTED = 1
    STATE_CLOSED = 2

    _STOP = object()

    def __init__(self, client, path):
        self._client = client
        self._root = TreeNode.make_root(self, path)
        self._state = self.STATE_LATENT
        self._outstanding_ops = 0
        self._is_initialized = False
        self._error_listeners = []
        self._event_listeners = []
        self._task_queue = client.handler.queue_impl()
        self._task_thread = None

    def start(self):
        """Starts the cache.

        The cache is not started automatically. You must call this method.

        After a cache started, all changes of subtree will be synchronized
        from the ZooKeeper server. Events will be fired for those activity.

        Don't forget to call :meth:`close` if a tree was started and you don't
        need it anymore, or you will leak the memory of cached nodes, even if
        you have released all references to the :class:`TreeCache` instance.
        Because there are so many callbacks that have been registered to the
        Kazoo client.

        See also :meth:`~TreeCache.listen`.

        .. note::

            This method is not thread safe.
        """
        if self._state == self.STATE_LATENT:
            self._state = self.STATE_STARTED
        elif self._state == self.STATE_CLOSED:
            raise KazooException("already closed")
        else:
            raise KazooException("already started")

        self._task_thread = self._client.handler.spawn(self._do_background)
        self._client.add_listener(self._session_watcher)
        self._client.ensure_path(self._root._path)

        if self._client.connected:
            # The on_created and other on_* methods must not be invoked outside
            # the background task. This is the key to keep concurrency safe
            # without lock.
            self._in_background(self._root.on_created)

    def close(self):
        """Closes the cache.

        A closed cache was detached from ZooKeeper's changes. And all nodes
        will be invalidated.

        Once a tree cache was closed, it could not be started again. You should
        only close a tree cache while you want to recycle it.

        .. note::

            This method is not thread safe.
        """
        if self._state == self.STATE_STARTED:
            self._state = self.STATE_CLOSED
            self._task_queue.put(self._STOP)
            self._client.remove_listener(self._session_watcher)
            with handle_exception(self._error_listeners):
                # We must invoke on_deleted outside background queue because:
                # 1. The background task has been stopped.
                # 2. The on_deleted on closed tree does not communicate with
                #    ZooKeeper actually.
                self._root.on_deleted()

    def listen(self, listener):
        """Registers a function to listen the cache events.

        The cache events are changes of local data. They are delivered from
        watching notifications in ZooKeeper session.

        This method can be use as a decorator.

        :param listener: A callable object which accepting a
                         :class:`~kazoo.recipe.cache.TreeEvent` instance as
                         its argument.
        """
        self._event_listeners.append(listener)
        return listener

    def listen_fault(self, listener):
        """Registers a function to listen the exceptions.

        It is possible to meet some exceptions during the cache running. You
        could specific handlers for them.

        This method can be use as a decorator.

        :param listener: A callable object which accepting an exception as its
                         argument.
        """
        self._error_listeners.append(listener)
        return listener

    def get_data(self, path, default=None):
        """Gets data of a node from cache.

        :param path: The absolute path string.
        :param default: The default value which will be returned if the node
                        does not exist.
        :raises ValueError: If the path is outside of this subtree.
        :returns: A :class:`~kazoo.recipe.cache.NodeData` instance.
        """
        node = self._find_node(path)
        return default if node is None else node._data

    def get_children(self, path, default=None):
        """Gets node children list from in-memory snapshot.

        :param path: The absolute path string.
        :param default: The default value which will be returned if the node
                        does not exist.
        :raises ValueError: If the path is outside of this subtree.
        :returns: The :class:`frozenset` which including children names.
        """
        node = self._find_node(path)
        return default if node is None else frozenset(node._children)

    def _find_node(self, path):
        if not path.startswith(self._root._path):
            raise ValueError("outside of tree")
        striped_path = path[len(self._root._path) :].strip("/")
        splited_path = [p for p in striped_path.split("/") if p]
        current_node = self._root
        for node_name in splited_path:
            if node_name not in current_node._children:
                return
            current_node = current_node._children[node_name]
        return current_node

    def _publish_event(self, event_type, event_data=None):
        event = TreeEvent.make(event_type, event_data)
        if self._state != self.STATE_CLOSED:
            logger.debug("public event: %r", event)
            self._in_background(self._do_publish_event, event)

    def _do_publish_event(self, event):
        for listener in self._event_listeners:
            with handle_exception(self._error_listeners):
                listener(event)

    def _in_background(self, func, *args, **kwargs):
        self._task_queue.put((func, args, kwargs))

    def _do_background(self):
        while True:
            with handle_exception(self._error_listeners):
                cb = self._task_queue.get()
                if cb is self._STOP:
                    break
                func, args, kwargs = cb
                func(*args, **kwargs)

                # release before possible idle
                del cb, func, args, kwargs

    def _session_watcher(self, state):
        if state == KazooState.SUSPENDED:
            self._publish_event(TreeEvent.CONNECTION_SUSPENDED)
        elif state == KazooState.CONNECTED:
            # The session watcher should not be blocked
            self._in_background(self._root.on_reconnected)
            self._publish_event(TreeEvent.CONNECTION_RECONNECTED)
        elif state == KazooState.LOST:
            self._is_initialized = False
            self._publish_event(TreeEvent.CONNECTION_LOST)


class TreeNode(object):
    """The tree node record.

    :param tree: A :class:`~kazoo.recipe.cache.TreeCache` instance.
    :param path: The path of current node.
    :param parent: The parent node reference. ``None`` for root node.
    """

    __slots__ = (
        "_tree",
        "_path",
        "_parent",
        "_depth",
        "_children",
        "_state",
        "_data",
    )

    STATE_PENDING = 0
    STATE_LIVE = 1
    STATE_DEAD = 2

    def __init__(self, tree, path, parent):
        self._tree = tree
        self._path = path
        self._parent = parent
        self._depth = parent._depth + 1 if parent else 0
        self._children = {}
        self._state = self.STATE_PENDING
        self._data = None

    @classmethod
    def make_root(cls, tree, path):
        return cls(tree, path, None)

    def on_reconnected(self):
        self._refresh()
        for child in self._children.values():
            child.on_reconnected()

    def on_created(self):
        self._refresh()

    def on_deleted(self):
        old_children, self._children = self._children, {}
        old_data, self._data = self._data, None

        for old_child in old_children.values():
            old_child.on_deleted()

        if self._tree._state == self._tree.STATE_CLOSED:
            self._reset_watchers()
            return

        old_state, self._state = self._state, self.STATE_DEAD
        if old_state == self.STATE_LIVE:
            self._publish_event(TreeEvent.NODE_REMOVED, old_data)

        if self._parent is None:
            self._call_client("exists", self._path)  # root node
        else:
            child = self._path[len(self._parent._path) + 1 :]
            if self._parent._children.get(child) is self:
                del self._parent._children[child]
                self._reset_watchers()

    def _publish_event(self, *args, **kwargs):
        return self._tree._publish_event(*args, **kwargs)

    def _reset_watchers(self):
        client = self._tree._client
        for _watchers in (client._data_watchers, client._child_watchers):
            _path = _prefix_root(client.chroot, self._path)
            _watcher = _watchers.get(_path, set())
            _watcher.discard(self._process_watch)

    def _refresh(self):
        self._refresh_data()
        self._refresh_children()

    def _refresh_data(self):
        self._call_client("get", self._path)

    def _refresh_children(self):
        # TODO max-depth checking support
        self._call_client("get_children", self._path)

    def _call_client(self, method_name, path):
        assert method_name in ("get", "get_children", "exists")
        self._tree._outstanding_ops += 1
        callback = functools.partial(
            self._tree._in_background, self._process_result, method_name, path
        )
        method = getattr(self._tree._client, method_name + "_async")
        method(path, watch=self._process_watch).rawlink(callback)

    def _process_watch(self, watched_event):
        logger.debug("process_watch: %r", watched_event)
        with handle_exception(self._tree._error_listeners):
            if watched_event.type == EventType.CREATED:
                assert self._parent is None, "unexpected CREATED on non-root"
                self.on_created()
            elif watched_event.type == EventType.DELETED:
                self.on_deleted()
            elif watched_event.type == EventType.CHANGED:
                self._refresh_data()
            elif watched_event.type == EventType.CHILD:
                self._refresh_children()

    def _process_result(self, method_name, path, result):
        logger.debug("process_result: %s %s", method_name, path)
        if method_name == "exists":
            assert self._parent is None, "unexpected EXISTS on non-root"
            # The result will be `None` if the node doesn't exist.
            if result.successful() and result.get() is not None:
                if self._state == self.STATE_DEAD:
                    self._state = self.STATE_PENDING
                self.on_created()
        elif method_name == "get_children":
            if result.successful():
                children = result.get()
                for child in sorted(children):
                    full_path = kazoo_join(path, child)
                    if child not in self._children:
                        node = TreeNode(self._tree, full_path, self)
                        self._children[child] = node
                        node.on_created()
            elif isinstance(result.exception, NoNodeError):
                self.on_deleted()
        elif method_name == "get":
            if result.successful():
                data, stat = result.get()
                old_data, self._data = (
                    self._data,
                    NodeData.make(path, data, stat),
                )
                old_state, self._state = self._state, self.STATE_LIVE
                if old_state == self.STATE_LIVE:
                    if old_data is None or old_data.stat.mzxid != stat.mzxid:
                        self._publish_event(TreeEvent.NODE_UPDATED, self._data)
                else:
                    self._publish_event(TreeEvent.NODE_ADDED, self._data)
            elif isinstance(result.exception, NoNodeError):
                self.on_deleted()
        else:  # pragma: no cover
            logger.warning("unknown operation %s", method_name)
            self._tree._outstanding_ops -= 1
            return

        self._tree._outstanding_ops -= 1
        if self._tree._outstanding_ops == 0 and not self._tree._is_initialized:
            self._tree._is_initialized = True
            self._publish_event(TreeEvent.INITIALIZED)


class TreeEvent(tuple):
    """The immutable event tuple of cache."""

    NODE_ADDED = 0
    NODE_UPDATED = 1
    NODE_REMOVED = 2
    CONNECTION_SUSPENDED = 3
    CONNECTION_RECONNECTED = 4
    CONNECTION_LOST = 5
    INITIALIZED = 6

    #: An enumerate integer to indicate event type.
    event_type = property(operator.itemgetter(0))

    #: A :class:`~kazoo.recipe.cache.NodeData` instance.
    event_data = property(operator.itemgetter(1))

    @classmethod
    def make(cls, event_type, event_data):
        """Creates a new TreeEvent tuple.

        :returns: A :class:`~kazoo.recipe.cache.TreeEvent` instance.
        """
        assert event_type in (
            cls.NODE_ADDED,
            cls.NODE_UPDATED,
            cls.NODE_REMOVED,
            cls.CONNECTION_SUSPENDED,
            cls.CONNECTION_RECONNECTED,
            cls.CONNECTION_LOST,
            cls.INITIALIZED,
        )
        return cls((event_type, event_data))


class NodeData(tuple):
    """The immutable node data tuple of cache."""

    #: The absolute path string of current node.
    path = property(operator.itemgetter(0))

    #: The bytes data of current node.
    data = property(operator.itemgetter(1))

    #: The stat information of current node.
    stat = property(operator.itemgetter(2))

    @classmethod
    def make(cls, path, data, stat):
        """Creates a new NodeData tuple.

        :returns: A :class:`~kazoo.recipe.cache.NodeData` instance.
        """
        return cls((path, data, stat))


@contextlib.contextmanager
def handle_exception(listeners):
    try:
        yield
    except Exception as e:
        logger.debug("processing error: %r", e)
        if listeners:
            for listener in listeners:
                try:
                    listener(e)
                except BaseException:  # pragma: no cover
                    logger.exception("Exception handling exception")  # oops
        else:
            logger.exception("No listener to process %r", e)



"""Zookeeper Barriers

:Maintainer: None
:Status: Unknown

"""
import os
import socket
import uuid

from kazoo.exceptions import KazooException, NoNodeError, NodeExistsError
from kazoo.protocol.states import EventType


class Barrier(object):
    """Kazoo Barrier

    Implements a barrier to block processing of a set of nodes until
    a condition is met at which point the nodes will be allowed to
    proceed. The barrier is in place if its node exists.

    .. warning::

        The :meth:`wait` function does not handle connection loss and
        may raise :exc:`~kazoo.exceptions.ConnectionLossException` if
        the connection is lost while waiting.

    """

    def __init__(self, client, path):
        """Create a Kazoo Barrier

        :param client: A :class:`~kazoo.client.KazooClient` instance.
        :param path: The barrier path to use.

        """
        self.client = client
        self.path = path

    def create(self):
        """Establish the barrier if it doesn't exist already"""
        self.client.retry(self.client.ensure_path, self.path)

    def remove(self):
        """Remove the barrier

        :returns: Whether the barrier actually needed to be removed.
        :rtype: bool

        """
        try:
            self.client.retry(self.client.delete, self.path)
            return True
        except NoNodeError:
            return False

    def wait(self, timeout=None):
        """Wait on the barrier to be cleared

        :returns: True if the barrier has been cleared, otherwise
                  False.
        :rtype: bool

        """
        cleared = self.client.handler.event_object()

        def wait_for_clear(event):
            if event.type == EventType.DELETED:
                cleared.set()

        exists = self.client.exists(self.path, watch=wait_for_clear)
        if not exists:
            return True

        cleared.wait(timeout)
        return cleared.is_set()


class DoubleBarrier(object):
    """Kazoo Double Barrier

    Double barriers are used to synchronize the beginning and end of
    a distributed task. The barrier blocks when entering it until all
    the members have joined, and blocks when leaving until all the
    members have left.

    .. note::

        You should register a listener for session loss as the process
        will no longer be part of the barrier once the session is
        gone. Connection losses will be retried with the default retry
        policy.

    """

    def __init__(self, client, path, num_clients, identifier=None):
        """Create a Double Barrier

        :param client: A :class:`~kazoo.client.KazooClient` instance.
        :param path: The barrier path to use.
        :param num_clients: How many clients must enter the barrier to
                            proceed.
        :type num_clients: int
        :param identifier: An identifier to use for this member of the
                           barrier when participating. Defaults to the
                           hostname + process id.

        """
        self.client = client
        self.path = path
        self.num_clients = num_clients
        self._identifier = identifier or "%s-%s" % (
            socket.getfqdn(),
            os.getpid(),
        )
        self.participating = False
        self.assured_path = False
        self.node_name = uuid.uuid4().hex
        self.create_path = self.path + "/" + self.node_name

    def enter(self):
        """Enter the barrier, blocks until all nodes have entered"""
        try:
            self.client.retry(self._inner_enter)
            self.participating = True
        except KazooException:
            # We failed to enter, best effort cleanup
            self._best_effort_cleanup()
            self.participating = False

    def _inner_enter(self):
        # make sure our barrier parent node exists
        if not self.assured_path:
            self.client.ensure_path(self.path)
            self.assured_path = True

        ready = self.client.handler.event_object()

        try:
            self.client.create(
                self.create_path,
                self._identifier.encode("utf-8"),
                ephemeral=True,
            )
        except NodeExistsError:
            pass

        def created(event):
            if event.type == EventType.CREATED:
                ready.set()

        self.client.exists(self.path + "/" + "ready", watch=created)

        children = self.client.get_children(self.path)

        if len(children) < self.num_clients:
            ready.wait()
        else:
            self.client.ensure_path(self.path + "/ready")
        return True

    def leave(self):
        """Leave the barrier, blocks until all nodes have left"""
        try:
            self.client.retry(self._inner_leave)
        except KazooException:  # pragma: nocover
            # Failed to cleanly leave
            self._best_effort_cleanup()
        self.participating = False

    def _inner_leave(self):
        # Delete the ready node if its around
        try:
            self.client.delete(self.path + "/ready")
        except NoNodeError:
            pass

        while True:
            children = self.client.get_children(self.path)
            if not children:
                return True

            if len(children) == 1 and children[0] == self.node_name:
                self.client.delete(self.create_path)
                return True

            children.sort()

            ready = self.client.handler.event_object()

            def deleted(event):
                if event.type == EventType.DELETED:
                    ready.set()

            if self.node_name == children[0]:
                # We're first, wait on the highest to leave
                if not self.client.exists(
                    self.path + "/" + children[-1], watch=deleted
                ):
                    continue

                ready.wait()
                continue

            # Delete our node
            self.client.delete(self.create_path)

            # Wait on the first
            if not self.client.exists(
                self.path + "/" + children[0], watch=deleted
            ):
                continue

            # Wait for the lowest to be deleted
            ready.wait()

    def _best_effort_cleanup(self):
        try:
            self.client.retry(self.client.delete, self.create_path)
        except NoNodeError:
            pass



import contextlib
import unittest

import pytest

from kazoo.client import KazooClient
from kazoo.handlers import utils
from kazoo.protocol import states as kazoo_states
from kazoo.tests import test_client
from kazoo.tests import test_lock
from kazoo.tests import util as test_util

try:
    import eventlet
    from eventlet.green import threading
    from kazoo.handlers import eventlet as eventlet_handler

    EVENTLET_HANDLER_AVAILABLE = True
except ImportError:
    EVENTLET_HANDLER_AVAILABLE = False


@contextlib.contextmanager
def start_stop_one(handler=None):
    if not handler:
        handler = eventlet_handler.SequentialEventletHandler()
    handler.start()
    try:
        yield handler
    finally:
        handler.stop()


class TestEventletHandler(unittest.TestCase):
    def setUp(self):
        if not EVENTLET_HANDLER_AVAILABLE:
            pytest.skip("eventlet handler not available.")
        super(TestEventletHandler, self).setUp()

    def test_started(self):
        with start_stop_one() as handler:
            assert handler.running is True
            assert len(handler._workers) != 0
        assert handler.running is False
        assert len(handler._workers) == 0

    def test_spawn(self):
        captures = []

        def cb():
            captures.append(1)

        with start_stop_one() as handler:
            handler.spawn(cb)

        assert len(captures) == 1

    def test_dispatch(self):
        captures = []

        def cb():
            captures.append(1)

        with start_stop_one() as handler:
            handler.dispatch_callback(kazoo_states.Callback("watch", cb, []))

        assert len(captures) == 1

    def test_async_link(self):
        captures = []

        def cb(handler):
            captures.append(handler)

        with start_stop_one() as handler:
            r = handler.async_result()
            r.rawlink(cb)
            r.set(2)

        assert len(captures) == 1
        assert r.get() == 2

    def test_timeout_raising(self):
        handler = eventlet_handler.SequentialEventletHandler()

        with pytest.raises(handler.timeout_exception):
            raise handler.timeout_exception("This is a timeout")

    def test_async_ok(self):
        captures = []

        def delayed():
            captures.append(1)
            return 1

        def after_delayed(handler):
            captures.append(handler)

        with start_stop_one() as handler:
            r = handler.async_result()
            r.rawlink(after_delayed)
            w = handler.spawn(utils.wrap(r)(delayed))
            w.join()

        assert len(captures) == 2
        assert captures[0] == 1
        assert r.get() == 1

    def test_get_with_no_block(self):
        handler = eventlet_handler.SequentialEventletHandler()

        with start_stop_one(handler):
            r = handler.async_result()

            with pytest.raises(handler.timeout_exception):
                r.get(block=False)
            r.set(1)
            assert r.get() == 1

    def test_async_exception(self):
        def broken():
            raise IOError("Failed")

        with start_stop_one() as handler:
            r = handler.async_result()
            w = handler.spawn(utils.wrap(r)(broken))
            w.join()

        assert r.successful() is False
        with pytest.raises(IOError):
            r.get()

    def test_huge_file_descriptor(self):
        try:
            import resource
        except ImportError:
            self.skipTest("resource module unavailable on this platform")
        from eventlet.green import socket
        from kazoo.handlers.utils import create_tcp_socket

        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (4096, 4096))
        except (ValueError, resource.error):
            self.skipTest("couldnt raise fd limit high enough")
        fd = 0
        socks = []
        while fd < 4000:
            sock = create_tcp_socket(socket)
            fd = sock.fileno()
            socks.append(sock)
        with start_stop_one() as h:
            h.start()
            h.select(socks, [], [], 0)
            h.stop()
        for sock in socks:
            sock.close()


class TestEventletClient(test_client.TestClient):
    def setUp(self):
        if not EVENTLET_HANDLER_AVAILABLE:
            pytest.skip("eventlet handler not available.")
        super(TestEventletClient, self).setUp()

    @staticmethod
    def make_event():
        return threading.Event()

    @staticmethod
    def make_condition():
        return threading.Condition()

    def _makeOne(self, *args):
        return eventlet_handler.SequentialEventletHandler(*args)

    def _get_client(self, **kwargs):
        kwargs["handler"] = self._makeOne()
        return KazooClient(self.hosts, **kwargs)


class TestEventletSemaphore(test_lock.TestSemaphore):
    def setUp(self):
        if not EVENTLET_HANDLER_AVAILABLE:
            pytest.skip("eventlet handler not available.")
        super(TestEventletSemaphore, self).setUp()

    @staticmethod
    def make_condition():
        return threading.Condition()

    @staticmethod
    def make_event():
        return threading.Event()

    @staticmethod
    def make_thread(*args, **kwargs):
        return threading.Thread(*args, **kwargs)

    def _makeOne(self, *args):
        return eventlet_handler.SequentialEventletHandler(*args)

    def _get_client(self, **kwargs):
        kwargs["handler"] = self._makeOne()
        c = KazooClient(self.hosts, **kwargs)
        try:
            self._clients.append(c)
        except AttributeError:
            self._client = [c]
        return c


class TestEventletLock(test_lock.KazooLockTests):
    def setUp(self):
        if not EVENTLET_HANDLER_AVAILABLE:
            pytest.skip("eventlet handler not available.")
        super(TestEventletLock, self).setUp()

    @staticmethod
    def make_condition():
        return threading.Condition()

    @staticmethod
    def make_event():
        return threading.Event()

    @staticmethod
    def make_thread(*args, **kwargs):
        return threading.Thread(*args, **kwargs)

    @staticmethod
    def make_wait():
        return test_util.Wait(getsleep=(lambda: eventlet.sleep))

    def _makeOne(self, *args):
        return eventlet_handler.SequentialEventletHandler(*args)

    def _get_client(self, **kwargs):
        kwargs["handler"] = self._makeOne()
        c = KazooClient(self.hosts, **kwargs)
        try:
            self._clients.append(c)
        except AttributeError:
            self._client = [c]
        return c



import os
import subprocess
import time

import pytest

from kazoo.testing import KazooTestHarness
from kazoo.exceptions import (
    AuthFailedError,
    NoAuthError,
)
from kazoo.tests.util import CI_ZK_VERSION


class TestLegacySASLDigestAuthentication(KazooTestHarness):
    def setUp(self):
        try:
            import puresasl  # NOQA
        except ImportError:
            pytest.skip("PureSASL not available.")

        os.environ["ZOOKEEPER_JAAS_AUTH"] = "digest"
        self.setup_zookeeper()

        if CI_ZK_VERSION:
            version = CI_ZK_VERSION
        else:
            version = self.client.server_version()
        if not version or version < (3, 4):
            pytest.skip("Must use Zookeeper 3.4 or above")

    def tearDown(self):
        self.teardown_zookeeper()
        os.environ.pop("ZOOKEEPER_JAAS_AUTH", None)

    def test_connect_sasl_auth(self):
        from kazoo.security import make_acl

        username = "jaasuser"
        password = "jaas_password"

        acl = make_acl("sasl", credential=username, all=True)

        sasl_auth = "%s:%s" % (username, password)
        client = self._get_client(auth_data=[("sasl", sasl_auth)])

        client.start()
        try:
            client.create("/1", acl=(acl,))
            # give ZK a chance to copy data to other node
            time.sleep(0.1)
            with pytest.raises(NoAuthError):
                self.client.get("/1")
        finally:
            client.delete("/1")
            client.stop()
            client.close()

    def test_invalid_sasl_auth(self):
        client = self._get_client(auth_data=[("sasl", "baduser:badpassword")])
        with pytest.raises(AuthFailedError):
            client.start()


class TestSASLDigestAuthentication(KazooTestHarness):
    def setUp(self):
        try:
            import puresasl  # NOQA
        except ImportError:
            pytest.skip("PureSASL not available.")

        os.environ["ZOOKEEPER_JAAS_AUTH"] = "digest"
        self.setup_zookeeper()

        if CI_ZK_VERSION:
            version = CI_ZK_VERSION
        else:
            version = self.client.server_version()
        if not version or version < (3, 4):
            pytest.skip("Must use Zookeeper 3.4 or above")

    def tearDown(self):
        self.teardown_zookeeper()
        os.environ.pop("ZOOKEEPER_JAAS_AUTH", None)

    def test_connect_sasl_auth(self):
        from kazoo.security import make_acl

        username = "jaasuser"
        password = "jaas_password"

        acl = make_acl("sasl", credential=username, all=True)

        client = self._get_client(
            sasl_options={
                "mechanism": "DIGEST-MD5",
                "username": username,
                "password": password,
            }
        )
        client.start()
        try:
            client.create("/1", acl=(acl,))
            # give ZK a chance to copy data to other node
            time.sleep(0.1)
            with pytest.raises(NoAuthError):
                self.client.get("/1")
        finally:
            client.delete("/1")
            client.stop()
            client.close()

    def test_invalid_sasl_auth(self):
        client = self._get_client(
            sasl_options={
                "mechanism": "DIGEST-MD5",
                "username": "baduser",
                "password": "badpassword",
            }
        )
        with pytest.raises(AuthFailedError):
            client.start()


class TestSASLGSSAPIAuthentication(KazooTestHarness):
    def setUp(self):
        try:
            import puresasl  # NOQA
        except ImportError:
            pytest.skip("PureSASL not available.")
        try:
            import kerberos  # NOQA
        except ImportError:
            pytest.skip("Kerberos support not available.")
        if not os.environ.get("KRB5_TEST_ENV"):
            pytest.skip("Test Kerberos environ not setup.")

        os.environ["ZOOKEEPER_JAAS_AUTH"] = "gssapi"
        self.setup_zookeeper()

        if CI_ZK_VERSION:
            version = CI_ZK_VERSION
        else:
            version = self.client.server_version()
        if not version or version < (3, 4):
            pytest.skip("Must use Zookeeper 3.4 or above")

    def tearDown(self):
        self.teardown_zookeeper()
        os.environ.pop("ZOOKEEPER_JAAS_AUTH", None)

    def test_connect_gssapi_auth(self):
        from kazoo.security import make_acl

        # Ensure we have a client ticket
        subprocess.check_call(
            [
                "kinit",
                "-kt",
                os.path.expandvars("${KRB5_TEST_ENV}/client.keytab"),
                "client",
            ]
        )

        acl = make_acl("sasl", credential="client@KAZOOTEST.ORG", all=True)

        client = self._get_client(sasl_options={"mechanism": "GSSAPI"})
        client.start()
        try:
            client.create("/1", acl=(acl,))
            # give ZK a chance to copy data to other node
            time.sleep(0.1)
            with pytest.raises(NoAuthError):
                self.client.get("/1")
        finally:
            client.delete("/1")
            client.stop()
            client.close()

    def test_invalid_gssapi_auth(self):
        # Request a post-datated ticket, so that it is currently invalid.
        subprocess.check_call(
            [
                "kinit",
                "-kt",
                os.path.expandvars("${KRB5_TEST_ENV}/client.keytab"),
                "-s",
                "30min",
                "client",
            ]
        )

        client = self._get_client(sasl_options={"mechanism": "GSSAPI"})
        with pytest.raises(AuthFailedError):
            client.start()



import uuid

from kazoo.testing import KazooTestCase


class KazooPartyTests(KazooTestCase):
    def setUp(self):
        super(KazooPartyTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_party(self):
        parties = [self.client.Party(self.path, "p%s" % i) for i in range(5)]

        one_party = parties[0]

        assert list(one_party) == []
        assert len(one_party) == 0

        participants = set()
        for party in parties:
            party.join()
            participants.add(party.data.decode("utf-8"))

            assert set(party) == participants
            assert len(party) == len(participants)

        for party in parties:
            party.leave()
            participants.remove(party.data.decode("utf-8"))

            assert set(party) == participants
            assert len(party) == len(participants)

    def test_party_reuse_node(self):
        party = self.client.Party(self.path, "p1")
        self.client.ensure_path(self.path)
        self.client.create(party.create_path)
        party.join()
        assert party.participating is True
        party.leave()
        assert party.participating is False
        assert len(party) == 0

    def test_party_vanishing_node(self):
        party = self.client.Party(self.path, "p1")
        party.join()
        assert party.participating is True
        self.client.delete(party.create_path)
        party.leave()
        assert party.participating is False
        assert len(party) == 0


class KazooShallowPartyTests(KazooTestCase):
    def setUp(self):
        super(KazooShallowPartyTests, self).setUp()
        self.path = "/" + uuid.uuid4().hex

    def test_party(self):
        parties = [
            self.client.ShallowParty(self.path, "p%s" % i) for i in range(5)
        ]

        one_party = parties[0]

        assert list(one_party) == []
        assert len(one_party) == 0

        participants = set()
        for party in parties:
            party.join()
            participants.add(party.data.decode("utf-8"))

            assert set(party) == participants
            assert len(party) == len(participants)

        for party in parties:
            party.leave()
            participants.remove(party.data.decode("utf-8"))

            assert set(party) == participants
            assert len(party) == len(participants)
